'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 29th March 2021 07:58:18 am
Last Modified: Friday, 25th August 2023 02:08:10 pm
'''
from copy import deepcopy
from typing import Iterator, List, Tuple, Optional
from warnings import warn
import os
import logging
import json
import warnings
import yaml

from mpi4py import MPI
import numpy as np
from obspy import Stream, UTCDateTime, Inventory, Trace
from tqdm import tqdm

from seismic.correlate.stream import CorrTrace, CorrStream
from seismic.correlate import preprocessing_td as pptd
from seismic.correlate import preprocessing_stream as ppst
from seismic.db.corr_hdf5 import CorrelationDataBase
from seismic.trace_data.waveform import Store_Client
from seismic.utils.fetch_func_from_str import func_from_str
from seismic.utils import miic_utils as mu


class Correlator(object):
    """
    Object to manage the actual Correlation (i.e., Green's function retrieval)
    for the database.
    """
    def __init__(self, store_client: Store_Client, options: dict or str):
        """
        Initiates the Correlator object. When executing
        :func:`~seismic.correlate.correlate.Correlator.pxcorr()`, it will
        actually compute the correlations and save them in an hdf5 file that
        can be handled using
        :class:`~seismic.db.corr_hdf5.CorrelationDataBase`.
        Data has to be preprocessed before calling this (i.e., the data already
        has to be given in an ASDF format). Consult
        :class:`~seismic.trace_data.preprocess.Preprocessor` for information on
        how to proceed with this.

        :param options: Dictionary containing all options for the correlation.
            Can also be a path to a yaml file containing all the keys required
            in options.
        :type options: dict or str
        """
        if isinstance(options, str):
            with open(options) as file:
                options = yaml.load(file, Loader=yaml.FullLoader)
        # init MPI
        self.comm = MPI.COMM_WORLD
        self.psize = self.comm.Get_size()
        self.rank = self.comm.Get_rank()
        # directories
        self.proj_dir = options['proj_dir']
        self.corr_dir = os.path.join(self.proj_dir, options['co']['subdir'])
        try:
            self.save_comps_separately = options['save_comps_separately']
        except KeyError:
            self.save_comps_separately = False
        logdir = os.path.join(self.proj_dir, options['log_subdir'])
        if self.rank == 0:
            os.makedirs(self.corr_dir, exist_ok=True)
            os.makedirs(logdir, exist_ok=True)

        # Logging - rank dependent
        if self.rank == 0:
            tstr = UTCDateTime.now().strftime('%Y-%m-%d-%H:%M')
        else:
            tstr = None
        tstr = self.comm.bcast(tstr, root=0)

        rankstr = str(self.rank).zfill(3)

        loglvl = mu.log_lvl[options['log_level'].upper()]
        self.logger = logging.getLogger("seismic.Correlator%s" % rankstr)
        self.logger.setLevel(loglvl)
        logging.captureWarnings(True)
        warnlog = logging.getLogger('py.warnings')
        fh = logging.FileHandler(os.path.join(logdir, 'correlate%srank%s' % (
            tstr, rankstr)))
        fh.setLevel(loglvl)
        self.logger.addHandler(fh)
        warnlog.addHandler(fh)
        fmt = logging.Formatter(
            fmt='%(asctime)s - %(levelname)s - %(message)s')
        fh.setFormatter(fmt)
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(fmt)
        self.logger.addHandler(consoleHandler)

        # Write the options dictionary to the log file
        if self.rank == 0:
            opt_dump = deepcopy(options)
            # json cannot write the UTCDateTime objects that might be in here
            for step in opt_dump['co']['preProcessing']:
                if 'stream_mask_at_utc' in step['function']:
                    startsstr = [
                        t.format_fissures() for t in step['args']['starts']]
                    step['args']['starts'] = startsstr
                    if 'ends' in step['args']:
                        endsstr = [
                            t.format_fissures() for t in step['args']['ends']]
                        step['args']['ends'] = endsstr
            with open(os.path.join(
                    logdir, 'params%s.txt' % tstr), 'w') as file:
                file.write(json.dumps(opt_dump, indent=1))

        self.options = options['co']

        # requested combis?
        if 'xcombinations' in self.options:
            self.rcombis = self.options['xcombinations']
            if self.rcombis == 'None':
                # cumbersome, but someone used it wrong so let's hardcode
                self.rcombis = None
        else:
            self.rcombis = None

        # find the available data
        network = options['net']['network']
        station = options['net']['station']

        # Store_Client
        self.store_client = store_client

        if isinstance(station, list) and len(station) == 1:
            station = station[0]
        if isinstance(network, list) and len(network) == 1:
            network = network[0]

        if (
            network == '*'
                and isinstance(station, str) and '*' not in station):
            raise ValueError(
                'Stations has to be either: \n'
                + '1. A list of the same length as the list of networks.\n'
                + '2. \'*\' That is, a wildcard (string).\n'
                + '3. A list and network is a string describing one '
                + 'station code.')
        elif isinstance(station, str) and isinstance(network, str):
            station = [[network, station]]
        elif station == '*' and isinstance(network, list):
            # This is most likely not thread-safe
            if self.rank == 0:
                station = []
                for net in network:
                    station.extend(store_client.get_available_stations(net))
            else:
                station = None
            station = self.comm.bcast(station, root=0)
        elif isinstance(network, list) and isinstance(station, list):
            if len(network) != len(station):
                raise ValueError(
                    'Stations has to be either: \n'
                    + '1. A list of the same length as the list of networks.\n'
                    + '2. \'*\' That is, a wildcard (string).\n'
                    + '3. A list and network is a string describing one '
                    + 'station code.')
            station = list([n, s] for n, s in zip(network, station))
        elif isinstance(station, list) and isinstance(network, str):
            for ii, stat in enumerate(station):
                station[ii] = [network, stat]
        else:
            raise ValueError(
                'Stations has to be either: \n'
                + '1. A list of the same length as the list of networks.\n'
                + '2. \'*\' That is, a wildcard (string).\n'
                + '3. A list and network is a string describing one '
                + 'station code.')
        if self.rank == 0:
            self.avail_raw_data = []
            for net, stat in station:
                self.avail_raw_data.extend(
                    self.store_client._translate_wildcards(net, stat))
            # make sure this only contains unique combinations
            # with several cores it added entries several times, don't know
            # why?
            self.avail_raw_data = np.unique(
                self.avail_raw_data, axis=0).tolist()
        else:
            self.avail_raw_data = None
        self.avail_raw_data = self.comm.bcast(
            self.avail_raw_data)
        self.station = np.unique(np.array([
            [d[0], d[1]] for d in self.avail_raw_data]), axis=0).tolist()
        self.logger.debug(
            'Fetching data from the following stations:\n%a' % [
                f'{n}.{s}' for n, s in self.station])

        self.sampling_rate = self.options['sampling_rate']

    def find_interstat_dist(self, dis: float):
        """
        Find stations in database with interstation distance smaller than
        dis.

        If no station inventories are available, they will be downloaded.

        :param dis: Find all Stations with distance less than `dis` [in m]
        :type dis: float

        .. note:: only the subset of the in ``params.yaml`` defined
            networks and stations will be queried.
        """
        if not self.options['combination_method'] == 'betweenStations':
            raise ValueError(
                'This function is only available if combination method '
                + 'is set to "betweenStations".')
        # list of requested combinations
        self.rcombis = []
        # Update the store clients invetory
        self.store_client.read_inventory()
        for ii, (n0, s0) in enumerate(self.station):
            inv0 = self.store_client.select_inventory_or_load_remote(n0, s0)
            for n1, s1 in self.station[ii:]:
                inv1 = self.store_client.select_inventory_or_load_remote(
                    n1, s1)
                if mu.filter_stat_dist(inv0, inv1, dis):
                    self.rcombis.append('%s-%s.%s-%s' % (n0, n1, s0, s1))

    def find_existing_times(self, tag: str, channel: str = '*') -> dict:
        """
        Returns the already existing starttimes in form of a dictionary (see
        below)

        :param tag: The tag that the waveforms are saved under
        :type tag: str
        :param channel: Channel Combination Code (e.g., CH0-CH1),
            wildcards accepted. Defaults to '*'
        :type channel: str, optional
        :return: Dictionary that is structured as in the example below
        :rtype: dict

        Examples
        --------
        >>> out_dict = my_correlator.find_existing_times('mytag', 'BHZ-BHH')
        >>> print(out_dict)
        {'NET0.STAT0': {
            'NET1.STAT1': {'BHZ-BHH': [%list of starttimes] ,
            'NET2.STAT2': {'BHZ-BHH':[%list of starttimes]}}}
        """
        netlist, statlist = list(zip(*self.station))
        netcombs, statcombs = compute_network_station_combinations(
            netlist, statlist, method=self.options['combination_method'],
            combis=self.rcombis)
        ex_dict = {}
        for nc, sc in zip(netcombs, statcombs):
            outf = os.path.join(
                self.corr_dir, '%s.%s.h5' % (nc, sc))
            if not os.path.isfile(outf):
                continue
            with CorrelationDataBase(
                    outf, corr_options=self.options, mode='r') as cdb:
                d = cdb.get_available_starttimes(nc, sc, tag, channel)
            s0, s1 = sc.split('-')
            n0, n1 = nc.split('-')
            ex_dict.setdefault('%s.%s' % (n0, s0), {})
            ex_dict['%s.%s' % (n0, s0)]['%s.%s' % (n1, s1)] = d
        return ex_dict

    def pxcorr(self):
        """
        Start the correlation with the parameters that were defined when
        initiating the object.
        """
        cst = CorrStream()
        if self.rank == 0:
            self.logger.debug('Reading Inventory files.')
        # Fetch station coordinates
        if self.rank == 0:
            try:
                inv = self.store_client.read_inventory()
            except Exception as e:
                if self.options['remove_response']:
                    raise FileNotFoundError(
                        'No response information could be found.'
                        + 'If you set remove_response to True, you will need'
                        + 'a station inventory.')
                logging.warning(e)
                warnings.warn(
                    'No Station Inventory found. Proceeding without.')
                inv = None
        else:
            inv = None
        inv = self.comm.bcast(inv, root=0)

        for st, write_flag in self._generate_data():
            cst.extend(self._pxcorr_inner(st, inv))
            if write_flag:
                self.logger.debug('Writing Correlations to file.')
                # Here, we can recombine the correlations for the read_len
                # size (i.e., stack)
                # Write correlations to HDF5
                if cst.count():
                    self._write(cst)
                    cst.clear()

        # write the remaining data
        if cst.count():
            self._write(cst)
            cst.clear()

    def _pxcorr_inner(self, st: Stream, inv: Inventory) -> CorrStream:
        """
        Inner loop of pxcorr. Don't call this function!
        """

        # We start out by moving the stream into a matrix
        self.logger.debug(
            'Converting Stream to Matrix')
        # put all the data into a single stream
        starttime = []
        npts = []
        for tr in st:
            starttime.append(tr.stats['starttime'])
            npts.append(tr.stats['npts'])
        npts = np.max(np.array(npts))

        A, st = st_to_np_array(st, npts)
        self.options.update(
            {'starttime': starttime,
                'sampling_rate': self.sampling_rate})
        self.logger.debug('Computing Cross-Correlations.')
        A, startlags = self._pxcorr_matrix(A)
        self.logger.debug('Converting Matrix to CorrStream.')
        # put trace into a stream
        cst = CorrStream()
        if A is None:
            # No new data
            return cst
        if self.rank == 0:
            for ii, (startlag, comb) in enumerate(
                    zip(startlags, self.options['combinations'])):
                endlag = startlag + len(A[ii, :])/self.options['sampling_rate']
                cst.append(
                    CorrTrace(
                        A[ii], header1=st[comb[0]].stats,
                        header2=st[comb[1]].stats, inv=inv, start_lag=startlag,
                        end_lag=endlag))
        else:
            cst = None
        cst = self.comm.bcast(cst, root=0)
        return cst

    def _write(self, cst):
        """
        Write correlation stream to files.

        :param cst: CorrStream containing the correlations
        :type cst: :class:`~seismic.correlate.stream.CorrStream`
        """
        if not cst.count():
            self.logger.debug('No new data written.')
            return

        # Make sure that each core writes to a different file
        if self.save_comps_separately:
            codelist = list(set(
                [f'{tr.stats.network}.{tr.stats.station}.{tr.stats.channel}'
                 for tr in cst]))
        else:
            codelist = list(set(
                [f'{tr.stats.network}.{tr.stats.station}' for tr in cst]))
        # Better if the same cores keep writing to the same files
        codelist.sort()
        # Decide which process writes to which station
        pmap = np.arange(len(codelist))*self.psize/len(codelist)
        pmap = pmap.astype(np.int32)
        ind = pmap == self.rank

        for code in np.array(codelist)[ind]:
            if self.save_comps_separately:
                net, stat, cha = code.split('.')
                outf = os.path.join(self.corr_dir, f'{net}.{stat}.{cha}.h5')
                cstselect = cst.select(
                    network=net, station=stat, channel=cha)
            else:
                net, stat = code.split('.')
                outf = os.path.join(self.corr_dir, f'{net}.{stat}.h5')
                cstselect = cst.select(network=net, station=stat)
            if self.options['subdivision']['recombine_subdivision']:
                stack = cstselect.stack(regard_location=False)
                stacktag = 'stack_%s' % str(self.options['read_len'])
            else:
                stack = None
            if self.options['subdivision']['delete_subdivision']:
                cstselect.clear()
            with CorrelationDataBase(
                    outf, corr_options=self.options) as cdb:
                if cstselect.count():
                    cdb.add_correlation(cstselect, 'subdivision')
                if stack is not None:
                    cdb.add_correlation(stack, stacktag)

    def _generate_data(self) -> Iterator[Tuple[Stream, bool]]:
        """
        Returns an Iterator that loops over each start and end time with the
        requested window length.

        Also will lead to the time windows being prolonged by the sum of
        the length of the tapered end, so that no data is lost. Always
        a hann taper so far. Taper length is 5% of the requested time
        window on each side.

        :yield: An obspy stream containing the time window x for all stations
        that were active during this time.
        :rtype: Iterator[Stream]
        """
        if self.rank == 0:
            # find already available times
            self.ex_dict = self.find_existing_times('subdivision')
            self.logger.info('Already existing data: %s' % str(self.ex_dict))
        else:
            self.ex_dict = None

        self.ex_dict = self.comm.bcast(self.ex_dict, root=0)

        if not self.ex_dict and self.options['preprocess_subdiv']:
            self.options['preprocess_subdiv'] = False
            if self.rank == 0:
                self.logger.warning(
                    'No existing data found.\nAutomatically setting '
                    'preprocess_subdiv to False to optimise performance.')

        # the time window that the loop will go over
        t0 = UTCDateTime(self.options['read_start']).timestamp
        t1 = UTCDateTime(self.options['read_end']).timestamp
        loop_window = np.arange(t0, t1, self.options['read_inc'])

        # Taper ends for the deconvolution
        if self.options['remove_response']:
            tl = 100
        else:
            tl = 0

        # Decide which process reads data from which station
        # Better than just letting one core read as this avoids having to
        # send very big chunks of data using MPI (MPI communication does
        # not support more than 2GB/comm operation)
        pmap = np.arange(len(self.avail_raw_data))*self.psize/len(
            self.avail_raw_data)
        pmap = pmap.astype(np.int32)
        ind = pmap == self.rank
        ind = np.arange(len(self.avail_raw_data))[ind]

        # Loop over read increments
        for t in tqdm(loop_window):
            write_flag = True  # Write length is same as read length
            startt = UTCDateTime(t)
            endt = startt + self.options['read_len']
            st = Stream()
            resp = Inventory()

            # loop over queried stations
            for net, stat, cha in np.array(self.avail_raw_data)[ind]:
                # Load data
                resp.extend(
                    self.store_client.inventory.select(
                        net, stat))
                stext = self.store_client._load_local(
                    net, stat, '*', cha, startt, endt, True, False)
                mu.get_valid_traces(stext)
                if stext is None or not len(stext):
                    # No data for this station to read
                    continue
                st.extend(stext)

            # Stream based preprocessing
            # Downsampling
            # 04/04/2023 Downsample before preprocessing for performance
            # Check sampling frequency
            sampling_rate = self.options['sampling_rate']
            # AA-Filter is done in this function as well
            st = mu.resample_or_decimate(st, sampling_rate)
            # The actual data in the mseeds was changed from int to float64
            # now,
            # Save some space by changing it back to 32 bit (most of the
            # digitizers work at 24 bit anyways)
            mu.stream_require_dtype(st, np.float32)

            if not self.options['preprocess_subdiv']:
                try:
                    self.logger.debug('Preprocessing stream...')
                    st = preprocess_stream(
                        st, self.store_client, resp, startt, endt, tl,
                        **self.options)
                except ValueError as e:
                    self.logger.error(
                        'Stream preprocessing failed for '
                        f'{st[0].stats.network}.{st[0].stats.station} and time'
                        f' {t}.\nThe Original Error Message was {e}.')
                    continue

            # Slice the stream in correlation length
            # -> Loop over correlation increments
            for ii, win in enumerate(generate_corr_inc(st, **self.options)):
                winstart = startt + ii*self.options['subdivision']['corr_inc']
                winend = winstart + self.options['subdivision']['corr_len']

                # Gather time windows from all stations to all cores
                winl = self.comm.allgather(win)
                win = Stream()
                for winp in winl:
                    win.extend(winp)
                win.sort(keys=['network', 'station', 'channel'])

                # Get correlation combinations
                if self.rank == 0:
                    self.logger.debug('Calculating combinations...')
                    self.options['combinations'] = calc_cross_combis(
                        win, self.ex_dict, self.options['combination_method'],
                        rcombis=self.rcombis)
                else:
                    self.options['combinations'] = None
                self.options['combinations'] = self.comm.bcast(
                    self.options['combinations'], root=0)

                if not len(self.options['combinations']):
                    # no new combinations for this time period
                    self.logger.info(
                        f'No new data for times {winstart}-{winend}')
                    continue
                # Remove traces that won't be accessed at all
                win_indices = np.arange(len(win))
                combindices = np.unique(
                    np.hstack(self.options['combinations']))
                popindices = np.flip(
                    np.setdiff1d(win_indices, combindices))
                for popi in popindices:
                    del win[popi]
                if len(popindices):
                    # now we have to recompute the combinations
                    if self.rank == 0:
                        self.logger.debug('removing redundant data.')
                        self.logger.debug('Recalculating combinations...')
                        self.options['combinations'] = calc_cross_combis(
                            win, self.ex_dict,
                            self.options['combination_method'],
                            rcombis=self.rcombis)
                    else:
                        self.options['combinations'] = None
                    self.options['combinations'] = self.comm.bcast(
                        self.options['combinations'], root=0)

                    if not len(self.options['combinations']):
                        # no new combinations for this time period
                        self.logger.info(
                            f'No new data for times {winstart}-{winend}')
                        continue
                # Stream based preprocessing
                if self.options['preprocess_subdiv']:
                    try:
                        win = preprocess_stream(
                            win, self.store_client, resp, winstart, winend,
                            tl, **self.options)
                    except ValueError as e:
                        self.logger.error(
                            'Stream preprocessing failed for '
                            f'{st[0].stats.network}.{st[0].stats.station}'
                            ' and time '
                            f'{t}.\nThe Original Error Message was {e}.')
                        continue
                    if self.rank == 0:
                        self.options['combinations'] = calc_cross_combis(
                            win, self.ex_dict,
                            self.options['combination_method'],
                            rcombis=self.rcombis)
                    else:
                        self.options['combinations'] = None
                    self.options['combinations'] = self.comm.bcast(
                        self.options['combinations'], root=0)

                if not len(win):
                    # no new combinations for this time period
                    self.logger.info(
                        f'No new data for times {winstart}-{winend}')
                    continue

                self.logger.debug('Working on correlation times %s-%s' % (
                    str(win[0].stats.starttime), str(win[0].stats.endtime)))
                yield win, write_flag
                write_flag = False

    def _pxcorr_matrix(self, A: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        # time domain processing
        # map of traces on processes
        ntrc = A.shape[0]
        pmap = np.arange(ntrc)*self.psize/ntrc
        # This step was not in the original but is necessary for it to work?
        # maybe a difference in an old python/np version?
        pmap = pmap.astype(np.int32)

        # indices for traces to be worked on by each process
        ind = pmap == self.rank

    ######################################
        corr_args = self.options['corr_args']
        # time domain pre-processing
        params = {}
        for key in list(corr_args.keys()):
            if 'Processing' not in key:
                params.update({key: corr_args[key]})
        params['sampling_rate'] = self.sampling_rate
        # The steps that aren't done before

        # nans from the masked parts are set to 0
        np.nan_to_num(A, copy=False)

        for proc in corr_args['TDpreProcessing']:
            func = func_from_str(proc['function'])
            A[ind, :] = func(A[ind, :], proc['args'], params)

        # zero-padding
        A = pptd.zeroPadding(A, {'type': 'avoidWrapFastLen'}, params)

        ######################################
        # FFT
        # Allocate space for rfft of data
        zmsize = A.shape

        # use next fast len instead?
        fftsize = zmsize[1]//2+1
        B = np.zeros((ntrc, fftsize), dtype=np.csingle)

        B[ind, :] = np.fft.rfft(A[ind, :], axis=1)

        freqs = np.fft.rfftfreq(zmsize[1], 1./self.sampling_rate)

        ######################################
        # frequency domain pre-processing
        params.update({'freqs': freqs})
        # Here, I will have to make sure to add all the functions to the module
        for proc in corr_args['FDpreProcessing']:

            # The big advantage of this rather lengthy code is that we can also
            # import any function that has been defined anywhere else (i.e,
            # not only within the miic framework)
            func = func_from_str(proc['function'])
            B[ind, :] = func(B[ind, :], proc['args'], params)

        ######################################
        # collect results
        self.comm.Allreduce(MPI.IN_PLACE, [B, MPI.FLOAT], op=MPI.SUM)

        ######################################
        # correlation
        csize = len(self.options['combinations'])
        irfftsize = (fftsize-1)*2
        sampleToSave = int(
            np.ceil(
                corr_args['lengthToSave'] * self.sampling_rate))
        C = np.zeros((csize, sampleToSave*2+1), dtype=np.float32)

        pmap = np.arange(csize)*self.psize/csize
        pmap = pmap.astype(np.int32)
        ind = pmap == self.rank
        ind = np.arange(csize)[ind]
        startlags = np.zeros(csize, dtype=np.float32)
        for ii in ind:
            # offset of starttimes in samples(just remove fractions of samples)
            offset = (
                self.options['starttime'][
                    self.options['combinations'][ii][0]] - self.options[
                        'starttime'][self.options['combinations'][ii][1]])
            if corr_args['center_correlation']:
                roffset = 0.
            else:
                # offset exceeding a fraction of integer
                roffset = np.fix(
                    offset * self.sampling_rate) / self.sampling_rate
            # faction of samples to be compenasated by shifting
            offset -= roffset
            # normalization factor of fft correlation
            if corr_args['normalize_correlation']:
                norm = (
                    np.sqrt(
                        2.*np.sum(B[self.options[
                            'combinations'][ii][0], :]
                            * B[self.options['combinations'][ii][0], :].conj())
                        - B[self.options['combinations'][ii][0], 0]**2)
                    * np.sqrt(
                        2.*np.sum(B[self.options[
                            'combinations'][ii][1], :]
                            * B[self.options['combinations'][ii][1], :].conj())
                        - B[self.options['combinations'][ii][1], 0]**2)
                    / irfftsize).real
            else:
                norm = 1.

            M = (
                B[self.options['combinations'][ii][0], :].conj()
                * B[self.options['combinations'][ii][1], :]
                * np.exp(1j * freqs * offset * 2 * np.pi))

            ######################################
            # frequency domain postProcessing
            #
            tmp = np.fft.irfft(M).real

            # cut the center and do fftshift
            C[ii, :] = np.concatenate(
                (tmp[-sampleToSave:], tmp[:sampleToSave+1]))/norm
            startlags[ii] = - sampleToSave / self.sampling_rate \
                - roffset

        ######################################
        # time domain postProcessing

        ######################################
        # collect results
        self.logger.debug('%s %s' % (C.shape, C.dtype))
        self.logger.debug('combis: %s' % (self.options['combinations']))

        self.comm.Allreduce(MPI.IN_PLACE, [C, MPI.FLOAT], op=MPI.SUM)
        self.comm.Allreduce(
            MPI.IN_PLACE, [startlags, MPI.FLOAT], op=MPI.SUM)

        return (C, startlags)


def st_to_np_array(st: Stream, npts: int) -> Tuple[np.ndarray, Stream]:
    """
    Converts an obspy stream to a matrix with the shape (npts, st.count()).
    Also returns the same stream but without the data arrays in tr.data.

    :param st: Input Stream
    :type st: Stream
    :param npts: Maximum number of samples per Trace
    :type npts: int
    :return: A stream and a matrix
    :rtype: np.ndarray
    """
    A = np.zeros((st.count(), npts), dtype=np.float32)
    for ii, tr in enumerate(st):
        A[ii, :tr.stats.npts] = tr.data
        del tr.data  # Not needed any more, just uses up RAM
    return A, st


def _compare_existing_data(ex_corr: dict, tr0: Trace, tr1: Trace) -> bool:
    # The actual starttime for the header is the later one of the two
    net0 = tr0.stats.network
    stat0 = tr0.stats.station
    cha0 = tr0.stats.channel
    net1 = tr1.stats.network
    stat1 = tr1.stats.station
    cha1 = tr1.stats.channel
    # Probably faster than checking a huge dict twice
    flip = ([net0, net1], [stat0, stat1], [cha0, cha1]) \
        != sort_comb_name_alphabetically(
        net0, stat0, net1, stat1, cha0, cha1)
    corr_start = max(tr0.stats.starttime, tr1.stats.starttime)
    try:
        if flip:
            return corr_start.format_fissures() in ex_corr[
                f'{net1}.{stat1}'][f'{net0}.{stat0}'][
                '%s-%s' % (
                    tr1.stats.channel, tr0.stats.channel)]
        else:
            return corr_start.format_fissures() in ex_corr[
                f'{net0}.{stat0}'][f'{net1}.{stat1}'][
                '%s-%s' % (
                    tr0.stats.channel, tr1.stats.channel)]
    except KeyError:
        return False


def calc_cross_combis(
    st: Stream, ex_corr: dict, method: str = 'betweenStations',
        rcombis: List[str] = None) -> list:
    """
    Calculate a list of all cross correlation combination
    of traces in the stream: i.e. all combination with two different
    stations involved.

    :param st: Stream holding the tracecs to be correlated
    :type st: :class:`~obspy.Stream`
    :param ex_corr: dict holding the correlations that already exist in db
    :type ex_corr: dict
    :type method: stringf
    :param method: Determines which traces of the strem are combined.
    :param rcombis: requested combinations, only works if
        `method==betweenStations`.
    :type rcombis: List[str] strings are in form net0-net1.stat0-stat1

        ``'betweenStations'``:
            Traces are combined if either their station or
            their network names are different.
        ``'betweenComponents'``:
            Traces are combined if their components (last
            letter of channel name) names are different and their station and
            network names are identical (single station cross-correlation).
        ``'autoComponents'``:
            Traces are combined only with themselves.
        ``'allSimpleCombinations'``:
            All Traces are combined once (onle one of
            (0,1) and (1,0))
        ``'allCombinations'``:
            All traces are combined in both orders ((0,1) and (1,0))
    """

    combis = []
    # sort alphabetically
    st.sort(keys=['network', 'station', 'channel'])
    if method == 'betweenStations':
        for ii, tr in enumerate(st):
            for jj in range(ii+1, len(st)):
                tr1 = st[jj]
                n = tr.stats.network
                n2 = tr1.stats.network
                s = tr.stats.station
                s2 = tr1.stats.station
                if n != n2 or s != s2:
                    # check first whether this combi is in dict
                    if _compare_existing_data(ex_corr, tr, tr1):
                        continue
                    if rcombis is not None and not any(all(
                        i0 in i1 for i0 in [
                            n, n2, s, s2]) for i1 in rcombis):
                        # If particular combis are requested, compute only
                        # those
                        continue
                    combis.append((ii, jj))
    elif method == 'betweenComponents':
        for ii, tr in enumerate(st):
            for jj in range(ii+1, len(st)):
                tr1 = st[jj]
                if ((tr.stats['network'] == tr1.stats['network'])
                    and (tr.stats['station'] == tr1.stats['station'])
                    and (
                        tr.stats['channel'][-1] != tr1.stats['channel'][-1])):
                    if _compare_existing_data(ex_corr, tr, tr1):
                        continue
                    combis.append((ii, jj))
    elif method == 'autoComponents':
        for ii, tr in enumerate(st):
            if _compare_existing_data(ex_corr, tr, tr):
                continue
            combis.append((ii, ii))
    elif method == 'allSimpleCombinations':
        for ii, tr in enumerate(st):
            for jj in range(ii, len(st)):
                tr1 = st[jj]
                if _compare_existing_data(ex_corr, tr, tr1):
                    continue
                combis.append((ii, jj))
    elif method == 'allCombinations':
        for ii, tr in enumerate(st):
            for jj, tr1 in enumerate(st):
                if _compare_existing_data(ex_corr, tr, tr1):
                    continue
                combis.append((ii, jj))
    else:
        raise ValueError("Method has to be one of ('betweenStations', "
                         "'betweenComponents', 'autoComponents', "
                         "'allSimpleCombinations' or 'allCombinations').")
    if not len(combis):
        warn('Method %s found no combinations.' % method)
    return combis


# All the rotations are still untested, should do that at some point

# def rotate_multi_corr_stream(st: Stream) -> Stream:
#     """Rotate a stream with full Greens tensor from ENZ to RTZ

#     Take a stream with numerous correlation traces and rotate the
#     combinations of ENZ components into combinations of RTZ components in
#     all nine components of the Green's tensor are present. If not all nine
#     components are present no trace for this station combination is returned.

#     :type st: obspy.stream
#     :param st: stream with data in ENZ system
#     :rtype: obspy.stream
#     :return: stream in the RTZ system
#     """

#     out_st = Stream()
#     while st:
#         tl = list(range(9))
#         tst = st.select(network=st[0].stats['network'],
#                         station=st[0].stats['station'])
#         cnt = 0
#         for ttr in tst:
#             if ttr.stats['channel'][2] == 'E':
#                 if ttr.stats['channel'][6] == 'E':
#                     tl[0] = ttr
#                     cnt += 1
#                 elif ttr.stats['channel'][6] == 'N':
#                     tl[1] = ttr
#                     cnt += 2
#                 elif ttr.stats['channel'][6] == 'Z':
#                     tl[2] = ttr
#                     cnt += 4
#             elif ttr.stats['channel'][2] == 'N':
#                 if ttr.stats['channel'][6] == 'E':
#                     tl[3] = ttr
#                     cnt += 8
#                 elif ttr.stats['channel'][6] == 'N':
#                     tl[4] = ttr
#                     cnt += 16
#                 elif ttr.stats['channel'][6] == 'Z':
#                     tl[5] = ttr
#                     cnt += 32
#             elif ttr.stats['channel'][2] == 'Z':
#                 if ttr.stats['channel'][6] == 'E':
#                     tl[6] = ttr
#                     cnt += 64
#                 elif ttr.stats['channel'][6] == 'N':
#                     tl[7] = ttr
#                     cnt += 128
#                 elif ttr.stats['channel'][6] == 'Z':
#                     tl[8] = ttr
#                     cnt += 256
#         if cnt == 2**9-1:
#             st0 = Stream()
#             for t in tl:
#                 st0.append(t)
#             st1 = _rotate_corr_stream(st0)
#             out_st += st1
#         elif cnt == 27:  # only horizontal component combinations present
#             st0 = Stream()
#             for t in [0, 1, 3, 4]:
#                 st0.append(tl[t])
#             st1 = _rotate_corr_stream_horizontal(st0)
#             out_st += st1
#         elif cnt == 283:  # horizontal combinations + ZZ
#             st0 = Stream()
#             for t in [0, 1, 3, 4]:
#                 st0.append(tl[t])
#             st1 = _rotate_corr_stream_horizontal(st0)
#             out_st += st1
#             out_st.append(tl[8])
#         for ttr in tst:
#             for ind, tr in enumerate(st):
#                 if ttr.id == tr.id:
#                     st.pop(ind)

#     return out_st


# def _rotate_corr_stream_horizontal(st: Stream) -> Stream:
#     """ Rotate traces in stream from the EE-EN-NE-NN system to
#     the RR-RT-TR-TT system. The letters give the component order
#     in the input and output streams. Input traces are assumed to be of same
#     length and simultaneously sampled.
#     """

#     # rotation angles
#     # phi1 : counter clockwise angle between E and R(towards second station)
#     # the leading -1 accounts fact that we rotate the coordinate system,
#     # not a vector
#     phi1 = - np.pi/180*(90-st[0].stats['sac']['az'])
#     # phi2 : counter clockwise angle between E and R(away from first station)
#     phi2 = - np.pi/180*(90-st[0].stats['sac']['baz']+180)

#     c1 = np.cos(phi1)
#     s1 = np.sin(phi1)
#     c2 = np.cos(phi2)
#     s2 = np.sin(phi2)

#     rt = Stream()
#     RR = st[0].copy()
#     RR.data = c1*c2*st[0].data - c1*s2*st[1].data - s1*c2*st[2].data +\
#         s1*s2*st[3].data
#     tcha = list(RR.stats['channel'])
#     tcha[2] = 'R'
#     tcha[6] = 'R'
#     RR.stats['channel'] = ''.join(tcha)
#     rt.append(RR)

#     RT = st[0].copy()
#     RT.data = c1*s2*st[0].data + c1*c2*st[1].data - s1*s2*st[2].data -\
#         s1*c2*st[3].data
#     tcha = list(RT.stats['channel'])
#     tcha[2] = 'R'
#     tcha[6] = 'T'
#     RT.stats['channel'] = ''.join(tcha)
#     rt.append(RT)

#     TR = st[0].copy()
#     TR.data = s1*c2*st[0].data - s1*s2*st[1].data + c1*c2*st[2].data -\
#         c1*s2*st[3].data
#     tcha = list(TR.stats['channel'])
#     tcha[2] = 'T'
#     tcha[6] = 'R'
#     TR.stats['channel'] = ''.join(tcha)
#     rt.append(TR)

#     TT = st[0].copy()
#     TT.data = s1*s2*st[0].data + s1*c2*st[1].data + c1*s2*st[2].data +\
#         c1*c2*st[3].data
#     tcha = list(TT.stats['channel'])
#     tcha[2] = 'T'
#     tcha[6] = 'T'
#     TT.stats['channel'] = ''.join(tcha)
#     rt.append(TT)

#     return rt


# def _rotate_corr_stream(st: Stream) -> Stream:
#     """ Rotate traces in stream from the EE-EN-EZ-NE-NN-NZ-ZE-ZN-ZZ system to
#     the RR-RT-RZ-TR-TT-TZ-ZR-ZT-ZZ system. The letters give the component
#     in the input and output streams. Input traces are assumed to be of same
#     length and simultaneously sampled.
#     """

#     # rotation angles
#     # phi1 : counter clockwise angle between E and R(towards second station)
#     # the leading -1 accounts fact that we rotate the coordinate system,
#     # not a vector
#     phi1 = - np.pi/180*(90-st[0].stats['sac']['az'])
#     # phi2 : counter clockwise angle between E and R(away from first station)
#     phi2 = - np.pi/180*(90-st[0].stats['sac']['baz']+180)

#     c1 = np.cos(phi1)
#     s1 = np.sin(phi1)
#     c2 = np.cos(phi2)
#     s2 = np.sin(phi2)

#     rtz = Stream()
#     RR = st[0].copy()
#     RR.data = c1*c2*st[0].data - c1*s2*st[1].data - s1*c2*st[3].data +\
#         s1*s2*st[4].data
#     tcha = list(RR.stats['channel'])
#     tcha[2] = 'R'
#     tcha[6] = 'R'
#     RR.stats['channel'] = ''.join(tcha)
#     rtz.append(RR)

#     RT = st[0].copy()
#     RT.data = c1*s2*st[0].data + c1*c2*st[1].data - s1*s2*st[3].data -\
#         s1*c2*st[4].data
#     tcha = list(RT.stats['channel'])
#     tcha[2] = 'R'
#     tcha[6] = 'T'
#     RT.stats['channel'] = ''.join(tcha)
#     rtz.append(RT)

#     RZ = st[0].copy()
#     RZ.data = c1*st[2].data - s1*st[5].data
#     tcha = list(RZ.stats['channel'])
#     tcha[2] = 'R'
#     tcha[6] = 'Z'
#     RZ.stats['channel'] = ''.join(tcha)
#     rtz.append(RZ)

#     TR = st[0].copy()
#     TR.data = s1*c2*st[0].data - s1*s2*st[1].data + c1*c2*st[3].data -\
#         c1*s2*st[4].data
#     tcha = list(TR.stats['channel'])
#     tcha[2] = 'T'
#     tcha[6] = 'R'
#     TR.stats['channel'] = ''.join(tcha)
#     rtz.append(TR)

#     TT = st[0].copy()
#     TT.data = s1*s2*st[0].data + s1*c2*st[1].data + c1*s2*st[3].data +\
#         c1*c2*st[4].data
#     tcha = list(TT.stats['channel'])
#     tcha[2] = 'T'
#     tcha[6] = 'T'
#     TT.stats['channel'] = ''.join(tcha)
#     rtz.append(TT)

#     TZ = st[0].copy()
#     TZ.data = s1*st[2].data + c1*st[5].data
#     tcha = list(TZ.stats['channel'])
#     tcha[2] = 'T'
#     tcha[6] = 'Z'
#     TZ.stats['channel'] = ''.join(tcha)
#     rtz.append(TZ)

#     ZR = st[0].copy()
#     ZR.data = c2*st[6].data - s2*st[7].data
#     tcha = list(ZR.stats['channel'])
#     tcha[2] = 'Z'
#     tcha[6] = 'R'
#     ZR.stats['channel'] = ''.join(tcha)
#     rtz.append(ZR)

#     ZT = st[0].copy()
#     ZT.data = s2*st[6].data + c2*st[7].tuple
#     ZT.stats['channel'] = ''.join(tcha)
#     rtz.append(ZT)

#     rtz.append(st[8].copy())

#     return rtz


def sort_comb_name_alphabetically(
    network1: str, station1: str, network2: str, station2: str,
    channel1: Optional[str] = '',
        channel2: Optional[str] = '') -> Tuple[
        list, list]:
    """
    Returns the alphabetically sorted network and station codes from the two
    station.

    :param network1: network code of first station
    :type network1: str
    :param station1: station code of first station
    :type station1: str
    :param network2: Network code of second station
    :type network2: str
    :param station2: Station Code of second Station
    :type station2: str
    :return: A tuple containing the list of the network codes sorted
        and the list of the station codes sorted.
    :rtype: Tuple[ list, list]

    Examples
    --------
    >>> net1 = 'IU'  # Network Code of first station
    >>> stat1 = 'HRV'  # Station Code of first station
    >>> net2 = 'XN'
    >>> stat2 = 'NEP06'
    >>> print(sort_comb_name_aphabetically(
            net1, stat1, net2, stat2))
    (['IU', 'XN'], ['HRV', 'NEP06'])
    >>> print(sort_comb_name_aphabetically(
            net2, stat2, net1, stat1))
    (['IU', 'XN'], ['HRV', 'NEP06'])
    >>> # Different combination
    >>> net1 = 'YP'  # Network Code of first station
    >>> stat1 = 'AB3'  # Station Code of first station
    >>> net2 = 'XN'
    >>> stat2 = 'NEP06'
    >>> print(sort_comb_name_aphabetically(
            net1, stat1, net2, stat2))
    (['XN', 'YP'], ['NEP06', 'AB3'])
    >>> # Different combination
    >>> net1 = 'XN'  # Network Code of first station
    >>> stat1 = 'NEP07'  # Station Code of first station
    >>> net2 = 'XN'
    >>> stat2 = 'NEP06'
    >>> print(sort_comb_name_aphabetically(
            net1, stat1, net2, stat2))
    (['XN', 'XN'], ['NEP06', 'NEP07'])
    """
    if not all([isinstance(arg, str) for arg in [
            network1, network2, station1, station2]]):
        raise TypeError('All arguments have to be strings.')
    sort1 = network1 + station1 + channel1
    sort2 = network2 + station2 + channel2
    sort = [sort1, sort2]
    sorted = sort.copy()
    sorted.sort()
    if sort == sorted:
        netcomb = [network1, network2]
        statcomb = [station1, station2]
        chacomb = [channel1, channel2]
    else:
        netcomb = [network2, network1]
        statcomb = [station2, station1]
        chacomb = [channel2, channel1]
    return netcomb, statcomb, chacomb


def compute_network_station_combinations(
    netlist: list, statlist: list,
    method: str = 'betweenStations', combis: List[str] = None) -> Tuple[
        list, list]:
    """
    Return the network and station codes of the correlations for the provided
    lists of networks and stations and the queried combination method.

    :param netlist: List of network codes
    :type netlist: list
    :param statlist: List of Station Codes
    :type statlist: list
    :param method: The combination method to use. Has to be one of the
        following: `betweenStation`, `betweenComponents`, `autoComponents`,
        `allSimpleCombinations`, or `allCombinations`,
        defaults to 'betweenStations'.
    :type method: str, optional
    :param combis: List of desired station combinations.
        Given as [net0-net1.stat0-stat1]. Optional.
    :type combis: List[str]
    :raises ValueError: for unkown combination methods.
    :return: A tuple containing the list of the correlation network code
        and the list of the correlation station code.
    :rtype: Tuple[list, list]
    """
    netcombs = []
    statcombs = []
    if method == 'betweenStations':
        for ii, (n, s) in enumerate(zip(netlist, statlist)):
            for jj in range(ii+1, len(netlist)):
                n2 = netlist[jj]
                s2 = statlist[jj]
                if n != n2 or s != s2:
                    nc, sc, _ = sort_comb_name_alphabetically(n, s, n2, s2)
                    # Check requested combinations
                    if (
                        combis is not None
                        and f'{nc[0]}-{nc[1]}.{sc[0]}-{sc[1]}' not in combis
                    ):
                        continue
                    netcombs.append('%s-%s' % (nc[0], nc[1]))
                    statcombs.append('%s-%s' % (sc[0], sc[1]))

    elif method == 'betweenComponents' or method == 'autoComponents':
        netcombs = [n+'-'+n for n in netlist]
        statcombs = [s+'-'+s for s in statlist]
    elif method == 'allSimpleCombinations':
        for ii, (n, s) in enumerate(zip(netlist, statlist)):
            for jj in range(ii, len(netlist)):
                n2 = netlist[jj]
                s2 = statlist[jj]
                nc, sc, _ = sort_comb_name_alphabetically(n, s, n2, s2)
                netcombs.append('%s-%s' % (nc[0], nc[1]))
                statcombs.append('%s-%s' % (sc[0], sc[1]))
    elif method == 'allCombinations':
        for n, s in zip(netlist, statlist):
            for n2, s2 in zip(netlist, statlist):
                nc, sc, _ = sort_comb_name_alphabetically(n, s, n2, s2)
                netcombs.append('%s-%s' % (nc[0], nc[1]))
                statcombs.append('%s-%s' % (sc[0], sc[1]))
    else:
        raise ValueError("Method has to be one of ('betweenStations', "
                         "'betweenComponents', 'autoComponents', "
                         "'allSimpleCombinations' or 'allCombinations').")
    return netcombs, statcombs


def preprocess_stream(
    st: Stream, store_client: Store_Client, inv: Inventory or None,
    startt: UTCDateTime, endt: UTCDateTime, taper_len: float,
    remove_response: bool, subdivision: dict,
    preProcessing: List[dict] = None,
        **kwargs) -> Stream:
    """
    Does the preprocessing on a per stream basis. Most of the parameters can be
    fed in by using the "yaml" dict as kwargs.

    :param st: Input Stream to be processed
    :type st: :class:`obspy.core.stream.Stream`
    :param store_client: Store Client for the database
    :type store_client: :class:`~seismic.trace_data.waveform.Store_Client`
    :param inv: Station response, can be None if ``remove_response=False``.
    :type inv: Inventory or None
    :param startt: Starttime that the stream should be clipped / padded to.
    :type startt: :class:`obspy.UTCDateTime`
    :param endt: Endtime that the stream should be clipped / padded to.
    :type endt: :class:`obspy.UTCDateTime`
    :param taper_len: If the instrument response is removed, one might want to
        taper to mitigate the acausal effects of the deconvolution. This is
        the length of such a taper in seconds.
    :type taper_len: float
    :param remove_response: Should the instrument response be removed?
    :type remove_response: bool
    :param subdivision: Dictionary holding information about the correlation
        lenghts and increments.
    :type subdivision: dict
    :param preProcessing: List holding information about the different external
        preprocessing functions to be applied, defaults to None
    :type preProcessing: List[dict], optional
    :raises ValueError: For sampling rates higher than the stream's native
        sampling rate (upsampling is not permitted).
    :return: The preprocessed stream.
    :rtype: :class:`obspy.core.stream.Stream`
    """
    if not st.count():
        return st
    # To deal with any nans/masks
    st = st.split()
    st.sort(keys=['starttime'])

    # Clip to these again to remove the taper
    old_starts = [deepcopy(tr.stats.starttime) for tr in st]
    old_ends = [deepcopy(tr.stats.endtime) for tr in st]

    if remove_response:
        # taper before instrument response removal
        if taper_len:
            st = ppst.cos_taper_st(st, taper_len, False, True)
        try:
            if inv:
                ninv = inv
                st.attach_response(ninv)
            st.remove_response(taper=False)  # Changed for testing purposes
        except ValueError:
            print('Station response not found ... loading from remote.')
            # missing station response
            ninv = store_client.rclient.get_stations(
                network=st[0].stats.network, station=st[0].stats.station,
                channel='*', level='response')
            st.attach_response(ninv)
            st.remove_response(taper=False)
            store_client._write_inventory(ninv)

    # Sometimes Z has reversed polarity
    if inv:
        try:
            mu.correct_polarity(st, inv)
        except Exception as e:
            print(e)

    mu.discard_short_traces(st, subdivision['corr_len']/20)

    if preProcessing:
        for procStep in preProcessing:
            func = func_from_str(procStep['function'])
            st = func(st, **procStep['args'])
    # Remove the artificial taper from earlier
    for tr, ostart, oend in zip(st, old_starts, old_ends):
        tr.trim(starttime=ostart, endtime=oend)
    st.merge()
    st.trim(startt, endt, pad=True)

    mu.discard_short_traces(st, subdivision['corr_len']/20)
    return st


def generate_corr_inc(
    st: Stream, subdivision: dict, read_len: int,
        **kwargs) -> Iterator[Stream]:
    """
    Subdivides the preprocessed streams into parts of equal length using
    the parameters ``cor_inc`` and ``cor_len`` in ``subdivision``.
    This function can be acessed by several processes in parallel

    :param st: The preprocessed input stream
    :type st: :class:`obspy.core.stream.Stream`
    :param subdivision: Dictionary holding the information about the
        correlation length and increment.
    :type subdivision: dict
    :param read_len: Length to be read from disk in seconds
    :type read_len: int
    :yield: Equal length windows, padded with nans / masked if data is missing.
    :rtype: Generator[Stream]
    """

    try:
        # Second loop to return the time window in correlation length
        for ii, win0 in enumerate(st.slide(
            subdivision['corr_len']-st[0].stats.delta, subdivision['corr_inc'],
                include_partial_windows=True)):

            # We use trim so the windows have the right time and
            # are filled with masked arrays for places without values
            starttrim = st[0].stats.starttime + ii*subdivision['corr_inc']
            endtrim = st[0].stats.starttime + ii*subdivision['corr_inc'] +\
                subdivision['corr_len']-st[0].stats.delta
            win = win0.trim(starttrim, endtrim, pad=True)
            mu.get_valid_traces(win)

            yield win

    except IndexError:
        # processes with no data end up here
        win = Stream()
        # A little dirty, but it has to go through an equally long loop
        # else this will cause a deadlock
        for _ in range(int(np.ceil(
                read_len/subdivision['corr_inc']))):
            yield win



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 15th November 2022 05:15:30 pm
Last Modified: Monday, 16th January 2023 11:14:37 am
'''

from typing import Tuple

import numpy as np

from seismic.correlate.stats import CorrStats


def corr_mat_trim(
    data: np.ndarray, stats: CorrStats, starttime: float,
        endtime: float) -> Tuple[np.ndarray, CorrStats]:
    """ Trim the correlation matrix to a given period.

    Trim the correlation matrix `corr_mat` to the period from `starttime` to
    `endtime` given in seconds from the zero position, so both can be
    positive and negative. If `starttime` and `endtime` are datetime.datetime
    objects they are taken as absolute start and endtime.

    :type data: np.ndarray
    :param corr_mat: correlation matrix to be trimmed
    :type starttime: float
    :param starttime: start time in seconds with respect to the zero position.
        Hence, this parameter describes a lag!
    :type endtime: float
    :param order: end time in seconds with respect to the zero position

    :rtype tuple: Tuple[np.ndarray, CorrStats]
    :return: trimmed correlation matrix and new stats object
    """

    # fetch indices of start and end
    start = int(
        np.floor((starttime-stats['start_lag'])*stats['sampling_rate']))
    end = int(np.floor((endtime-stats['start_lag'])*stats['sampling_rate']))

    # check range
    if start < 0:
        print('Error: starttime before beginning of trace. Data not changed')
        return data, stats
    if end >= stats['npts']:
        print('Error: endtime after end of trace. Data not changed')
        return data, stats

    # select requested part from matrix
    # +1 is to include the last sample
    if len(data.shape) == 1:
        data = data[start:end+1]
        stats['npts'] = len(data)
    else:
        data = data[:, start:end + 1]
        stats['npts'] = data.shape[1]

    # set starttime, endtime and npts of the new stats
    stats['start_lag'] = starttime

    return data, stats



from seismic.correlate.correlate import Correlator # NOQA
from seismic.correlate.stream import CorrTrace, CorrStream # NOQA
from seismic.db.corr_hdf5 import CorrelationDataBase # NOQA
from seismic.monitor.dv import read_dv, DV # NOQA
from seismic.monitor.monitor import Monitor # NOQA
from seismic.trace_data.waveform import Store_Client # NOQA


'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Thursday, 18th February 2021 02:30:02 pm
Last Modified: Friday, 25th August 2023 02:37:05 pm
'''

import fnmatch
import os
import datetime
import glob
from typing import List, Tuple, Iterator
import warnings
import re

import numpy as np
from obspy.clients.fdsn import Client as rClient
from obspy.clients.fdsn.header import FDSNNoDataException
from obspy.clients.filesystem.sds import Client as lClient
from obspy import read_inventory, UTCDateTime, read, Stream, Inventory
from obspy.clients.fdsn.mass_downloader import RectangularDomain, \
    Restrictions, MassDownloader

from seismic.utils.raw_analysis import spct_series_welch


SDS_FMTSTR = os.path.join(
    "{year}", "{network}", "{station}", "{channel}.{sds_type}",
    "{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}")


class Store_Client(object):
    """
    Client for request and local storage of waveform data

    Request client that stores downloaded data for later local retrieval.
    When reading stored data the client reads from the local copy of the data.
    Inventory data is stored in the folder `inventory` and attached to data
    that is read.
    """

    def __init__(self, Client: rClient, path: str, read_only: bool = False):
        """
        Initialize the client

        :type Client: obspy request client that supports the 'get_waveform'
            method
        :param Client: obspy request client that supports the 'get_waveform'
            method
        :type path: str
        :param path: path of the store's sds root directory
        :type read_only: Bool
        :param read_only: If True the local archive is not extended.
        """
        assert os.path.isdir(path), "{} is not a directory".format(path)
        self.fileborder_seconds = 30
        self.fileborder_samples = 5000
        self.sds_root = os.path.join(path, 'mseed')
        if not read_only:
            os.makedirs(self.sds_root, exist_ok=True)
        self.inv_dir = os.path.join(path, "inventory")
        if os.path.isdir(self.inv_dir) and os.listdir(self.inv_dir):
            self.inventory = self.read_inventory()
        else:
            self.inventory = Inventory()
        if not read_only:
            os.makedirs(self.inv_dir, exist_ok=True)
        self.sds_type = "D"
        self.lclient = lClient(
            self.sds_root, sds_type=self.sds_type, format="MSEED",
            fileborder_seconds=self.fileborder_seconds,
            fileborder_samples=self.fileborder_samples)
        self.rclient = Client
        self.read_only = read_only

    def download_waveforms_mdl(
        self, starttime: UTCDateTime, endtime: UTCDateTime,
        clients: list or None = None, minlat: float or None = None,
        maxlat: float or None = None, minlon: float or None = None,
        maxlon: float or None = None, network: str or None = None,
        station: str or None = None, location: str or None = None,
            channel: str or None = None):
        # Initialise MassDownloader

        domain = RectangularDomain(
            minlatitude=minlat, maxlatitude=maxlat, minlongitude=minlon,
            maxlongitude=maxlon)

        restrictions = Restrictions(
            starttime=starttime,
            endtime=endtime,
            # Chunk it to have one file per day.
            chunklength_in_sec=86400,
            # Considering the enormous amount of data associated with
            # continuous requests, you might want to limit the data based on
            # SEED identifiers. If the location code is specified, the location
            # priority list is not used; the same is true for the channel
            # argument and priority list.
            network=network, station=station, location=location,
            channel=channel,
            # The typical use case for such a data set are noise correlations
            # where gaps are dealt with at a later stage.
            reject_channels_with_gaps=False,
            # Same is true with the minimum length. All data might be useful.
            minimum_length=0.0,
            # Guard against the same station having different names.
            minimum_interstation_distance_in_m=100.0,
            sanitize=False)

        mdl = MassDownloader(providers=clients)

        mdl.download(
            domain, restrictions, mseed_storage=self._get_mseed_storage,
            stationxml_storage=self.inv_dir)
        # update inventory
        self.read_inventory()

    def get_waveforms(
        self, network: str, station: str, location: str, channel: str,
        starttime: UTCDateTime, endtime: UTCDateTime,
            attach_response: bool = True, _check_times: bool = True) -> Stream:
        assert ~('?' in network+station+location+channel), \
            "Only single channel requests supported."
        assert ~('*' in network+station+location+channel), \
            "Only single channel requests supported."
        # try to load from local disk
        st = self._load_local(network, station, location, channel, starttime,
                              endtime, attach_response, _check_times)
        if st is None:
            st = self._load_remote(
                network, station, location, channel, starttime, endtime,
                attach_response)
        return st

    def get_available_stations(self, network: str = None) -> list:
        """
        Returns a list of stations for which raw data is available.
        Very similar to the
        :func:`~obspy.clients.filesystem.sds.get_all_stations()` method with
        the difference that this one allows to filter for particular networks.

        :param network: Only return stations from this network.
            ``network==None`` is same as a wildcard. Defaults to None.
        :type network: str or None, optional
        :return: List of network and station codes in the form:
            `[[net0,stat0], [net1,stat1]]`.
        :rtype: list
        """
        # If no network is defined use all
        network = network or '*'
        oslist = glob.glob(
            os.path.join(self.sds_root, '????', network, '*'))
        statlist = []
        for path in oslist:
            if not isinstance(eval(path.split(os.path.sep)[-3]), int):
                continue
            # Add all network and station combinations to list
            code = path.split(os.path.sep)[-2:]
            if code not in statlist:
                statlist.append(code)
        if not statlist:
            raise FileNotFoundError('No Data Available.')
        return statlist

    def _get_mseed_storage(
        self, network: str, station: str, location: str,
        channel: str, starttime: UTCDateTime,
            endtime: UTCDateTime) -> bool or str:

        # Returning True means that neither the data nor the StationXML file

        # will be downloaded.

        filename = SDS_FMTSTR.format(
            network=network, station=station, location=location,
            channel=channel, year=starttime.year, doy=starttime.julday,
            sds_type=self.sds_type)
        outf = os.path.join(self.sds_root, filename)

        if os.path.isfile(outf):
            return True

        # If a string is returned the file will be saved in that location.

        return outf

    def _get_times(self, network: str, station: str) -> Tuple[
            UTCDateTime, UTCDateTime]:
        """
        Return earliest and latest available time for a certain station.

        :param network: network code
        :type network: str
        :param station: station code
        :type station: str
        :return: (UTCDateTime, UTCDateTime)
        :rtype: tuple
        """
        dirlist = glob.glob(os.path.join(self.sds_root, '*', network, station))
        dirlist2 = [
            os.path.basename(
                os.path.dirname(os.path.dirname(i))) for i in dirlist]
        # We only want the folders with years
        l0 = fnmatch.filter(dirlist2, '1???')
        l1 = fnmatch.filter(dirlist2, '2???')
        del dirlist2
        dirlist = l0 + l1
        if not dirlist:
            warnings.warn(
                'Station %s.%s not in database' % (network, station),
                UserWarning)
            return (None, None)
        dirlist.sort()  # sort by time

        try:
            starttime = get_day_in_folder(
                self.sds_root, dirlist, network, station, '*', 'start')
            endtime = get_day_in_folder(
                self.sds_root, dirlist, network, station, '*', 'end')
        except FileNotFoundError as e:
            warnings.warn(e, UserWarning)
            return (None, None)
        return (starttime, endtime)

    def _translate_wildcards(self, network: str, station: str) -> List[
            List[str]]:
        """
        Look up network and station names in database, so that wildcards
        can be estimated into number of stations.

        :param network: network string, may contain wildcards
        :type network: str
        :param station: station string, may contain wildcards
        :type station: str
        :return: _description_
        :rtype: _type_
        """
        dirlist = glob.glob(
            os.path.join(self.sds_root, '*', network, station, '*'))
        nets = [os.path.basename(
            os.path.dirname(os.path.dirname(i))) for i in dirlist]
        stats = [os.path.basename(os.path.dirname(i)) for i in dirlist]
        chans = [os.path.basename(i).split('.')[0] for i in dirlist]
        # Create one nested list
        return [[nets[i], stats[i], chans[i]] for i in range(len(dirlist))]

    def _load_remote(
        self, network: str, station: str, location: str, channel: str,
        starttime: UTCDateTime, endtime: UTCDateTime,
            attach_response: bool) -> Stream:
        """
        Load data from remote resouce
        """
        print("Loading remotely.")
        try:
            st = self.rclient.get_waveforms(
                network, station, location, channel, starttime, endtime)
        except FDSNNoDataException as e:
            print(e)
            return Stream()
        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                st.attach_response(self.inventory)
            except Warning:
                print("Updating inventory.")
                ninv = self.rclient.get_stations(
                    network=network, station=station, channel=channel,
                    starttime=starttime, endtime=endtime, level='response')
                if attach_response:
                    st.attach_response(ninv)
                self._write_inventory(ninv)
        if ((len(st) > 0) and not self.read_only):
            self._write_local_data(st)
        return st

    def _load_local(
        self, network: str, station: str, location: str, channel: str,
        starttime: UTCDateTime, endtime: UTCDateTime, attach_response: bool,
            _check_times: bool) -> Stream or None:
        """
        Read data from local SDS structure.
        """
        # print("Loading locally ... ", end='')
        st = self.lclient.get_waveforms(
            network, station, location, channel, starttime, endtime)
        # Making sure that the order is correct for the next bit to work
        st.sort(['starttime'])
        if _check_times and (
            len(st) == 0 or (
                starttime < (st[0].stats.starttime-st[0].stats.delta) or (
                    endtime > st[-1].stats.endtime+st[-1].stats.delta))):
            # print("Failed")
            return None
        if attach_response:
            try:
                st.attach_response(self.inventory)
            except Exception:
                pass
        # print("Success")
        return st

    def read_inventory(self) -> Inventory:
        """
        Reads an up-to-date version of the corresponding Station Inventory.

        :return: Station Inventory
        :rtype: Inventory
        """
        self.inventory = read_inventory(os.path.join(self.inv_dir, '*.xml'))
        return self.inventory

    def select_inventory_or_load_remote(
            self, network: str, station: str) -> Inventory:
        """
        Checks whether the response for the provided station is in
        self.inventory if not the response will be fetched from remote.

        :param network: Network code
        :type network: str
        :param station: Station Code
        :type station: str
        :return: An Obspy Inventory object holding the response for the
            requested station.
        :rtype: Inventory
        """
        inv = self.inventory.select(network=network, station=station)
        if not len(inv):
            print('Station response not found ... loading from remote.')
            inv = self.rclient.get_stations(
                network=network, station=station,
                channel='*', level='response')
            self._write_inventory(inv)
        return inv

    def _write_local_data(self, st):
        """
        Write stream to local SDS structure.
        """
        # write the data
        for tr in st:
            starttime = tr.stats.starttime
            while starttime < tr.stats.endtime:
                endtime = UTCDateTime(
                    starttime.year, starttime.month, starttime.day) + 86400*1.5
                endtime = UTCDateTime(endtime.year, endtime.month, endtime.day)
                ttr = tr.copy().trim(starttime, endtime, nearest_sample=False)
                self._sds_write(ttr, starttime.year, starttime.julday)
                starttime = endtime

    def _sds_write(self, tr, year, julday):
        """
        Write max 1 day long trace to SDS structure.
        """
        filename = SDS_FMTSTR.format(
            network=tr.stats.network, station=tr.stats.station,
            location=tr.stats.location, channel=tr.stats.channel,
            year=year, doy=julday, sds_type=self.sds_type)
        filename = os.path.join(self.sds_root, filename)
        if os.path.exists(filename):
            rst = read(filename)
            rst.append(tr)
            rst.merge()
        else:
            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
            rst = tr
        rst.write(filename, format='MSEED')

    def _write_inventory(self, ninv: Inventory):
        """write the inventory information"""
        # Save inventory by station name, like this no old information
        # can be deleted
        if self.inventory is not None:
            self.inventory += ninv
        else:
            self.inventory = ninv
        fname = os.path.join(
            self.inv_dir, f'{ninv[0].code}.{ninv[0][0].code}.xml')
        self.inventory.write(fname, format="STATIONXML", validate=True)

    def _generate_time_windows(
        self, network: str, station: str, channel: str, starttime: UTCDateTime,
            endtime: UTCDateTime, increment: int = 86400) -> Iterator[Stream]:
        """
        Generates time windows with the requested increment from the requested
        station.

        :param network: network code
        :type network: str
        :param station: station code
        :type station: str
        :param channel: Channel Code
        :type channel: str
        :param starttime: Starttime of the first window
        :type starttime: UTCDateTime
        :param endtime: Endtime of the last window
        :type endtime: UTCDateTime
        :param increment: increment and window length between the windows,
            defaults to 86400
        :type increment: int, optional
        :yield: Stream with every window
        :rtype: Iterator[Stream]
        """
        starttimes = starttime.timestamp + np.arange(
            0, endtime.timestamp-starttime.timestamp, increment)
        for start in starttimes:
            end = UTCDateTime(start + increment)
            st = self._load_local(
                network, station, '*', channel, UTCDateTime(start), end,
                attach_response=True, _check_times=False)
            yield st

    def compute_spectrogram(
        self, network: str, station: str, channel: str, starttime: UTCDateTime,
        endtime: UTCDateTime, win_len: int, freq_max: float = 25,
            read_increment: int = 86400) -> Tuple[
                np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes a time series of spectrograms for the requested station and
        channel.

        # Enter plotting function here later
        .. seealso:: Use function :func:`~seismic.plot` to generate

        .. note::
            MPI support since version 0.4.2. Then the result is only
            available on rank 0. All other ranks return None. (to save RAM)

        :param network: network code
        :type network: str
        :param station: station code
        :type station: str
        :param channel: channel code
        :type channel: str
        :param starttime: starttime of the first window
        :type starttime: UTCDateTime
        :param endtime: endtime of the last window
        :type endtime: UTCDateTime
        :param win_len: Length of each time window in seconds
        :type win_len: int
        :param freq_max: Maximum frequency to compute, defaults to 25.
            Also determines the Nyquist frequency.
        :type freq_max: float, optional
        :param read_increment: Increment to read data in - does not
            influence the final result but has to be >= win_len,
            defaults to 86400
        :type read_increment: int, optional
        :return: Frequency vector, time vector and spectrogram matrix
        :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]
        """
        if read_increment < win_len:
            raise ValueError(
                'read_increment must be >= win_len, got {} and {}'.format(
                    read_increment, win_len))
        data_gen = self._generate_time_windows(
            network, station, channel, starttime, endtime, read_increment)
        return spct_series_welch(data_gen, win_len, freq_max)


class FS_Client(object):
    """
    Request Client for reading MSEED files from file system
    """
    def __init__(self, fs='SDS_archive'):
        """
        Initialize the client

        Refer to :class:`read_from_filesystem` for documentation
        """
        self.fs = fs

    def get_waveforms(
        self, network: str, station: str, location: str, channel: str,
        starttime: UTCDateTime, endtime: UTCDateTime, trim: bool = True,
            debug: bool = False, **kwargs) -> Stream:
        """
        Read data from the local file system.
        """
        ID = network+'.'+station+'.'+location+'.'+channel
        if isinstance(starttime, UTCDateTime):
            starttime = starttime.datetime
        if isinstance(endtime, UTCDateTime):
            endtime = endtime.datetime
        st = read_from_filesystem(
            ID, starttime, endtime, self.fs, trim=trim, debug=debug)
        return st


IDformat = ['%NET', '%net', '%STA', '%sta', '%LOC', '%loc', '%CHA', '%cha']


def read_from_filesystem(
    ID: str, starttime: datetime.datetime, endtime: datetime.datetime,
    fs: str = 'SDS_archive', trim: bool = True,
        debug: bool = False) -> Stream:
    """
    Read data from a filesystem

    Function to read miniSEED data from a given file structure by specifying
    ID and time interval plus a description of the file system structure.

    :param ID: seedID of the channel to read (NET.STA.LOC.CHA)
    :type ID: string
    :param starttime: start time
    :type starttime: datetime.datetime
    :param endtime: end time
    :type endtime: datetime.datetime
    :param fs: file structure descriptor
    :type fs: list
    :param trim: switch for trimming of the stream
    :type trim: bool
    :param debug: print debugging information
    :type debug: bool

    :rtype: :class:`obspy.Stream`
    :return: data stream of requested data

    If the switch for trimming is False the whole stream in the files is
    returned.

    **File structure descriptor**
    fs is a list of strings or other lists indicating the elements in the
    file structure. Each item of this list is translated in one level of the
    directory structure. The first element is a string indicating the
    base_directory. The following elements can be strings to indicate
    one of the following:

    - %X as defined by datetime.strftime indicating an element of t
        the time. e.g. %H
    - %NET: network name or %net for lower case network name
    - %STA: station name or %sta for lower case
    - %CHA: channel name or %cha for lower case
    - %LOC: location or %loc for lower case location code
    - string with out %

    The format strings are replaced either with an element of the starttime
    if they correspond to a datetime specifyer or with the respective part
    of the seedID. A string withouta % sign is not changed. If more than one
    format string is required in one directory level the need to be
    separated within a sublist.

    A format string for the ID can be followed by a pair of braces including
    two strings that will be used to replace the first string with the
    second. This can be used if the filename contains part of the ID in a
    different form.

    If fs is a single string it is interpreted as the base directory
    'SDSdir' of a SeisComP Data Structure (SDS) with TYPE fixed to D
    `<SDSdir>/Year/NET/STA/CHAN.TYPE/NET.STA.LOC.CHAN.TYPE.YEAR.DAY`
    This usage should be equivalent to `obspy.clients.filesystem.sds` client.


    :Example:

        Example for a station 'GJE' in network 'HEJSL' with channel 'BHZ' and
        location '00' with the start time 2010-12-24_11:36:30 and
        ``fs = ['base_dir','%Y','%b','%NET,['%j','_','%STA'','_T_',\
            "%CHA('BH','')", '.mseed']]``
        will be translated in a linux filename
        ``base_dir/2010/Nov/HEJSL/317_GJE_T_Z.mseed``

    .. note::

        If the data contain traces of different channels in the same file with
        different start and endtimes the routine will not work properly when a
        period spans multiple files.

    """

    # check input
    assert type(starttime) is datetime.datetime, \
        'starttime is not a datetime.datetime object: %s is type %s' % \
        (starttime, type(starttime))
    assert type(endtime) is datetime.datetime, \
        'endtime is not a datetime.datetime object: %s is type %s' % \
        (endtime, type(endtime))
    # check if fs is a string and set fs to SDS structure
    if isinstance(fs, str):
        fs = [fs, '%Y', '%NET', '%STA', ['%CHA', '.D'], [
            '%NET', '.', '%STA', '.', '%LOC', '.', '%CHA', '.D.', '%Y', '.',
            '%j']]
    # translate file structure string
    fpattern = _current_filepattern(ID, starttime, fs)
    if debug:
        print('Searching for files matching: %s\n at time %s\n' %
              (fpattern, starttime))
    st = _read_filepattern(fpattern, starttime, endtime, trim, debug)

    # if trace starts too late have a look in the previous section
    if (len(st) == 0) or (
            (st[0].stats.starttime-st[0].stats.delta).datetime > starttime):
        fpattern, _ = _adjacent_filepattern(ID, starttime, fs, -1)
        if debug:
            print('Searching for files matching: %s\n at time %s\n' %
                  (fpattern, starttime))
        st += _read_filepattern(fpattern, starttime, endtime, trim, debug)
        st.merge()
    thistime = starttime
    while ((len(st) == 0) or (st[0].stats.endtime.datetime < endtime)) & (
            thistime < endtime):
        fpattern, thistime = _adjacent_filepattern(ID, thistime, fs, 1)
        if debug:
            print('Searching for files matching: %s\n at time %s\n' %
                  (fpattern, thistime))
        if thistime == starttime:
            break
        st += _read_filepattern(fpattern, starttime, endtime, trim, debug)
        st.merge()
    if trim:
        st.trim(starttime=UTCDateTime(starttime), endtime=UTCDateTime(endtime))
    if debug:
        print('Following IDs are in the stream: ')
        for tr in st:
            print(tr.id)
        print('Selecting %s' % ID)
    st = st.select(id=ID)
    return st


def _read_filepattern(
    fpattern: str, starttime: datetime.datetime, endtime: datetime.datetime,
        trim: bool, debug: bool) -> Stream:
    """Read a stream from files whose names match a given pattern.
    """
    flist = glob.glob(fpattern)
    starttimes = []
    endtimes = []
    # first only read the header information
    for fname in flist:
        st = read(fname, headonly=True)
        starttimes.append(st[0].stats.starttime.datetime)
        endtimes.append(st[-1].stats.endtime.datetime)
    # now read the stream from the files that contain the period
    if debug:
        print('Matching files:\n')
        for (f, start, end) in zip(flist, starttimes, endtimes):
            print('%s from %s to %s\n' % (f, start, end))
    st = Stream()
    for ind, fname in enumerate(flist):
        if (starttimes[ind] < endtime) and (endtimes[ind] > starttime):
            if trim:
                st += read(
                    fname, starttime=UTCDateTime(starttime),
                    endtime=UTCDateTime(endtime))
            else:
                st += read(fname)
    try:
        st.merge()
    except Exception:
        print("Error merging traces for requested period!")
        st = Stream()
    return st


def _adjacent_filepattern(
    ID: str, starttime: datetime.datetime, fs: str,
        inc: int) -> Tuple[str, datetime.datetime]:
    """Return the file name that contains the data sequence prior to the one
    that contains the given time for a given file structure and ID.

    :param inc: either 1 for following of -1 for previous period
    :type inc: int
    """
    assert ((inc == 1) or (inc == -1)), " inc must either be 1 or -1"
    fname = ''
    flag = 0
    # find earlier time by turning back the time by one increment of the
    # last time indicator in the fs
    for part in fs[-1::-1]:
        if not isinstance(part, list):
            part = [part]
        for tpart in part[-1::-1]:
            if (not ((('(' in tpart) and (')' in tpart))
                or (tpart in IDformat))
                    and ('%' in tpart) and (flag == 0)):
                flag = 1
                if tpart in ['%H', '%I']:
                    thistime = starttime + inc * datetime.timedelta(hours=1)
                elif tpart == '%p':
                    thistime = starttime + inc * datetime.timedelta(hours=12)
                elif tpart in ['%a', '%A', '%w', '%d', '%j', '%-j']:
                    thistime = starttime + inc * datetime.timedelta(days=1)
                elif tpart in ['%U', '%W']:
                    thistime = starttime + inc * datetime.timedelta(days=7)
                elif tpart in ['%b', '%B', '%m']:
                    if starttime.month + inc == 0:
                        thistime = datetime.datetime(starttime.year-1,
                                                     12,
                                                     starttime.day,
                                                     starttime.hour,
                                                     starttime.minute,
                                                     starttime.second,
                                                     starttime.microsecond)
                    elif starttime.month + inc == 13:
                        thistime = datetime.datetime(starttime.year+1,
                                                     1,
                                                     starttime.day,
                                                     starttime.hour,
                                                     starttime.minute,
                                                     starttime.second,
                                                     starttime.microsecond)
                    else:
                        thistime = datetime.datetime(starttime.year,
                                                     starttime.month + inc,
                                                     starttime.day,
                                                     starttime.hour,
                                                     starttime.minute,
                                                     starttime.second,
                                                     starttime.microsecond)
                elif tpart in ['%y', '%Y']:
                    thistime = datetime.datetime(
                        starttime.year - inc, starttime.month, starttime.day,
                        starttime.hour, starttime.minute, starttime.second,
                        starttime.microsecond)
    fname = _current_filepattern(ID, thistime, fs)
    return fname, thistime


def _current_filepattern(
        ID: str, starttime: datetime.datetime, fs: str) -> str:
    """Return the file name that contains the data sequence that contains
    the given time for a given file structure and ID.
    """
    fname = ''
    for part in fs:
        if not isinstance(part, list):
            part = [part]
        fpartname = ''
        for tpart in part:
            fpartname += _fs_translate(tpart, ID, starttime)
        fname = os.path.join(fname, fpartname)
    return fname


def _fs_translate(part: str, ID: str, starttime: datetime.datetime) -> str:
    """Translate part of the file structure descriptor.
    """
    IDlist = ID.split('.')
    if ('(' in part) and (')' in part):
        trans = re.search('(.*?)', part).group(0)
    else:
        trans = None
    # in case there is something to translate remove from the filepart
    if trans:
        part = part.replace(trans, '')
    # if there is no %-sign it is a fixed string
    if '%' not in part:
        res = part
    # in case it belongs to the ID replace it with the respective ID-part
    if part in IDformat:
        idx = IDformat.index(part)
        res = IDlist[int(idx/2)]
        if idx % 2 != 0:  # idx is odd and IDformat is lowercase
            res = res.lower()
    # otherwise it must be part of the date string
    else:
        res = starttime.strftime(part)
    # replace if nesseccary
    if trans:
        transl = trans[1:-1].split(',')
        assert len(transl) == 2, "%s is not valid for replacement" % trans
        res = res.replace(
            transl[0].replace("'", ""), transl[1].replace("'", ""))
    return res


def get_day_in_folder(
    root: str, dirlist: List[str], network: str, station: str,
        channel: str, type: str) -> UTCDateTime:
    """
    Assuming that mseed files are by day (i.e., one file per day), this
    function will return the earliest or the latest day available (depending
    upon the argument passed as ``type``).

    :param root: The path to the sds root folder
    :type root: str
    :param dirlist: alphabetically sorted list of available years
    :type dirlist: List[str]
    :param network: The queried network code
    :type network: str
    :param station: The queried station code
    :type station: str
    :param channel: The queried channel code (wildcards allowed)
    :type channel: str
    :param type: either ``start`` or ``end``
    :type type: str
    :raises NotImplementedError: Unknown argument for type
    :raises FileNotFoundError: No files in db
    :return: The earliest starttime or latest endtime in utc
    :rtype: UTCDateTime
    """
    if type == 'start':
        i0 = 0
        ii = 1
    elif type == 'end':
        i0 = -1
        ii = -1
    else:
        raise NotImplementedError('Type has to be either start or end.')
    julday = None
    while not julday:
        # if the folder is empty julday will stay False
        year = dirlist[i0]
        julday = [i.split('.')[-1] for i in glob.glob(
            os.path.join(root, year, network, station, channel, '*'))]
        i0 += ii
    if not julday:
        raise FileNotFoundError(
            'Station %s.%s not in database' % (network, station))
    # make sure that 200 comes after 3, but before 300
    julday.sort(key=lambda li: (len(li), li))
    if type == 'start':
        return UTCDateTime(year=int(year), julday=julday[0])
    if type == 'end':
        # +24*3600 because else it's the starttime of latest day
        return UTCDateTime(year=int(year), julday=int(julday[-1])) + 24*3600



'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Friday, 9th April 2021 04:07:06 pm
Last Modified: Monday, 16th January 2023 11:13:58 am
'''


from typing import Callable


def func_from_str(string: str) -> Callable:
    """
    Given a full string in the form 'module.submodule.function'. This function
    imports said function and returns it, so that it can be called. The actual
    function can be located in any module.

    :param string: string in the form module.submodule.function
    :type string: str
    :return: The imported function ready to be executed
    :rtype: function
    """
    if not isinstance(string, str):
        raise TypeError('The input has to be a string!')
    splitl = string.split('.')
    funcname = splitl[-1]
    modname = '.'.join(splitl[:-1])
    module = __import__(modname, fromlist=['object'])
    func = getattr(module, funcname)
    return func



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 19th July 2021 11:37:54 am
Last Modified: Monday, 21st August 2023 12:21:53 pm
'''
import os
from typing import Tuple, Optional, List
import datetime

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.dates import date2num
import numpy as np
from obspy.core.utcdatetime import UTCDateTime
from obspy import Trace, Stream

from seismic.plot.plot_utils import set_mpl_params, remove_topright, remove_all


def plot_ctr(
    corr: Trace, tlim: list or tuple = None, ax: plt.Axes = None,
        outputdir: str = None, clean: bool = False) -> plt.Axes:
    set_mpl_params()

    # Get figure/axes dimensions
    if ax is None:
        width, height = 10, 2.5
        fig = plt.figure(figsize=(width, height))
        ax = plt.gca()
        axtmp = None
    else:
        fig = plt.gcf()
        bbox = ax.get_window_extent().transformed(
            fig.dpi_scale_trans.inverted())
        width, height = bbox.width, bbox.height
        axtmp = ax

    # The ratio ensures that the text
    # is perfectly distanced from top left/right corner
    ratio = width/height
    times = corr.times()
    ydata = corr.data

    # Plot stuff into axes
    ax.plot(times, ydata, 'k', lw=1)

    # Set limits
    if tlim is None:
        ax.set_xlim(times[0], times[-1])
    else:
        ax.set_xlim(tlim)

    ax.set_ylim([-1, 1])

    # Removes top/right axes spines. If you want the whole thing, comment
    # or remove
    remove_topright()

    # Plot RF only
    if clean:
        remove_all()
    else:
        ax.set_xlabel("Lag Time [s]")
        ax.set_ylabel("Correlation")
        # Start time in station stack does not make sense
        text = '%s - %s\n%s' % (
            corr.stats.corr_start.strftime('%Y/%m/%d %H:%M'),
            corr.stats.corr_end.strftime('%Y/%m/%d %H:%M'),
            corr.get_id())
        # corr.stats.corr_start.isoformat(sep=" ") + "\n" + corr.get_id()
        ax.text(0.995, 1.0-0.005*ratio, text, transform=ax.transAxes,
                horizontalalignment="right", verticalalignment="top")

    # Only use tight layout if not part of plot.
    if axtmp is None:
        plt.tight_layout()

    # Hide minor ticks on y-axis
    plt.tick_params(axis='y', which='minor', length=0)

    # Outout the receiver function as pdf using
    # its station name and starttime
    if outputdir is not None:
        filename = os.path.join(
            outputdir, corr.get_id() + "_"
            + corr.stats.starttime.strftime('%Y%m%dT%H%M%S')
            + ".pdf")
        plt.savefig(filename, format="pdf")
    return ax


def plot_cst(
    cst: Stream, sort_by: str = 'corr_start',
    timelimits: list or tuple or None = None,
    ylimits: list or tuple or None = None, scalingfactor: float = 2.0,
    ax: plt.Axes = None, linewidth: float = 0.25,
    outputfile: str or None = None, title: str or None = None,
    type: str = 'heatmap', cmap: str = 'seismic', vmin: float = None,
        vmax: float = None, **kwargs):
    """
    Creates a section or heat plot of all correlations in this stream.

    :param cst: Input CorrelationStream
    :type cst: :class:`~seismic.correlate.stream.CorrStream`
    :param sort_by: Which parameter to plot against. Can be either
        ``corr_start`` or ``distance``, defaults to 'corr_start'.
    :type sort_by: str, optional
    :param timelimits: xlimits (lag time) in seconds, defaults to None
    :type timelimits: Tuple[float, float], optional
    :param ylimits: limits for Y-axis (either a :class:`datetime.datetime`
        or float in km (if plotted against distance)), defaults to None.
    :type ylimits: Tuple[float, float], optional
    :param scalingfactor: Which factor to scale the Correlations with. Play
        around with this if you want to make amplitudes bigger or smaller,
        defaults to None (automatically chosen).
    :type scalingfactor: float, optional
    :param ax: Plot in existing axes? Defaults to None
    :type ax: plt.Axes, optional
    :param linewidth: Width of the lines to plot, defaults to 0.25
    :type linewidth: float, optional
    :param outputfile: Save the plot? defaults to None
    :type outputfile: str, optional
    :param title: Title of the plot, defaults to None
    :type title: str, optional
    :param type: can be set to 'heat' for a heatmap or 'section' for
        a wiggle type plot. Default to `heat`
    :type type: str, optional
    :param cmap: Decides about colormap if type == 'heatmap'.
        Defaults to 'inferno'.
    :type cmap: str, optional
    :return: returns the axes object.
    :rtype: `matplotlib.pyplot.Axes`

    """
    set_mpl_params()

    scalingfactor = scalingfactor or 1

    # Create figure if no axes is specified
    if ax is None:
        plt.figure(figsize=(8, 6))
        ax = plt.axes()  # zorder=9999999
    if sort_by == 'distance':
        sort_by = 'dist'
    cst.sort(keys=[sort_by, 'network', 'station', 'channel'])

    if not cst.count():
        raise KeyError('No Correlations with the desired properties found.')

    # Plot traces
    if sort_by == 'corr_start':
        scalingfactor *= 1e5
        if type == 'section':
            times = sect_plot_corr_start(cst, ax, scalingfactor, linewidth)
        elif type == 'heatmap':
            times = heat_plot_corr_start(
                cst, ax, cmap=cmap, vmin=vmin, vmax=vmax)
        else:
            raise NotImplementedError(
                'Unknown or not implemented plot type %s.' % type)
    elif sort_by == 'dist':
        if type == 'section':
            scalingfactor *= 4
            times = sect_plot_dist(
                cst, ax, scalingfactor, linewidth, **kwargs)
        elif type == 'heatmap':
            times = heat_plot_corr_dist(
                cst, ax, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    else:
        raise NotImplementedError('Unknown sorting method %s.' % sort_by)

    plt.tick_params(axis='y', which='minor', length=0)

    # Set limits
    if ylimits:
        plt.ylim(ylimits)

    if timelimits is None:
        plt.xlim(times[0], times[-1])
    else:
        plt.xlim(timelimits)

    plt.xlabel(r"$\tau$ [s]")

    plt.title(title)

    # Set output directory
    if outputfile is not None:
        plt.savefig(outputfile, dpi=300, transparent=True)
    return ax


def plot_corr_bulk(
    corr_bulk,
    timelimits: Optional[Tuple[datetime.datetime, datetime.datetime]] = None,
    ylimits: Optional[Tuple[datetime.datetime, datetime.datetime]] = None,
    clim: Optional[Tuple[float, float]] = None,
    plot_colorbar: bool = False, outputfile: Optional[str] = None,
    title: Optional[str] = None,
        ax: Optional[plt.Axes] = None) -> plt.Axes:
    """
    Plots a :class:`~seismic.correlate.stream.CorrBulk` object.

    :param corr_bulk: The CorrBulk to plot.
    :type corr_bulk: :class:`~seismic.correlate.stream.CorrBulk`
    :param timelimits: Limits time axis, defaults to None
    :type timelimits: Optional[Tuple[datetime.datetime, datetime.datetime]],
        optional
    :param ylimits: Limits of y-axis, defaults to None
    :type ylimits: Optional[Tuple[datetime.datetime, datetime.datetime]],
        optional
    :param clim: Limits of Colobar, defaults to None
    :type clim: Optional[Tuple[float, float]], optional
    :param plot_colorbar: add colorbar to plot, defaults to False
    :type plot_colorbar: bool, optional
    :param outputfile: save file to, defaults to None
    :type outputfile: Optional[str], optional
    :param title: Title, defaults to None
    :type title: Optional[str], optional
    :param ax: Axis to plot into, defaults to None
    :type ax: Optional[plt.Axes], optional
    :return: The current axis
    :rtype: plt.Axes
    """
    set_mpl_params()

    # Create figure if no axes is specified
    if ax is None:
        plt.figure(figsize=(8, 6))
        ax = plt.axes()  # zorder=9999999

    extent = [
        corr_bulk.stats.start_lag,
        corr_bulk.stats.end_lag,
        date2num(corr_bulk.stats.corr_start[0].datetime),
        date2num(corr_bulk.stats.corr_start[-1].datetime)]
    im = ax.imshow(corr_bulk.data, extent=extent, aspect='auto')
    ax.yaxis_date()
    ax.figure.autofmt_xdate(rotation=45)
    ax.set_xlabel(r'$\{tau}$ [s]')

    # Set limits
    if ylimits:
        ax.set_ylim(date2num(ylimits[0]), date2num(ylimits[1]))
    if timelimits:
        ax.set_xlim(timelimits)
    if title:
        ax.set_title(title)
    if clim:
        im.set_clim(clim)
    if plot_colorbar:
        plt.colorbar(im, orientation='vertical')
    if outputfile:
        plt.savefig(outputfile)
    else:
        plt.show()
    return ax


def sect_plot_corr_start(
    cst: Stream, ax: plt.Axes, scalingfactor: float,
        linewidth: float) -> np.ndarray:
    for ii, ctr in enumerate(cst):
        ydata = ctr.data
        times = ctr.times()

        ytmp = ctr.stats['corr_start'].timestamp

        ctmp = ydata * scalingfactor + ytmp
        ctmp = np.array([UTCDateTime(t).datetime for t in ctmp])

        ax.plot(times, ctmp, 'k', lw=linewidth, zorder=-ii + 0.1)

    ax.yaxis.set_major_locator(mpl.dates.AutoDateLocator())

    ax.yaxis.set_major_formatter(mpl.dates.DateFormatter('%d %h %y'))
    ax.invert_yaxis()
    return times


def heat_plot_corr_start(
        cst: Stream, ax: plt.Axes, cmap: str, vmin: float, vmax: float):
    data = np.empty((cst.count(), cst[0].stats.npts))
    # y grid
    y = []
    for ii, ctr in enumerate(cst):
        data[ii, :] = ctr.data
        y.append(ctr.stats['corr_start'].datetime)
        times = ctr.times()
    ds = plt.pcolormesh(
        times, np.array(y), data, shading='auto', cmap=cmap, vmin=vmin,
        vmax=vmax)
    plt.colorbar(
        ds, label='Correlation Coefficient', shrink=.6,
        orientation='horizontal')
    ax.yaxis.set_major_locator(mpl.dates.AutoDateLocator())
    ax.yaxis.set_major_formatter(mpl.dates.DateFormatter('%d %h %y'))
    ax.invert_yaxis()
    return times


def sect_plot_dist(
    cst: Stream, ax: plt.Axes, scalingfactor: float,
    linewidth: float, plot_reference_v: bool = False,
        ref_v: List[float] = [1, 2, 3]) -> np.ndarray:
    # Sort by azimuth. W on left side of plot
    for ii, ctr in enumerate(cst):
        ydata = ctr.data
        times = ctr.times()
        if ctr.stats.az >= 180:
            times = -1*times
        ytmp = ydata * scalingfactor + ctr.stats.dist

        ax.plot(times, ytmp, 'k', lw=linewidth, zorder=-ii + 0.1)
    if plot_reference_v:
        yref = np.array([ax.get_ylim()[1], 0, ax.get_ylim()[1]])
        [plt.plot(
            [-yref[0]/rv, 0, yref[0]/rv], yref, linewidth=.5, zorder=0,
            label=f'{rv}'+r'$\frac{km}{s}$') for rv in ref_v]
        plt.legend()
    plt.ylabel(r"Distance [km]")
    # Set label locations.
    # step = round((cst[-1].stats.dist - cst[0].stats.dist)/10000)
    # plt.yticks(np.arange(0, ctr.stats.dist/1000+step, step, dtype=int))
    plt.yticks(np.linspace(0, ctr.stats.dist, 10, dtype=int))
    return times


def heat_plot_corr_dist(
    cst: Stream, ax: plt.Axes, cmap: str, vmin: float, vmax: float,
        plot_reference_v: bool = False, ref_v: List[float] = [1, 2, 3]):
    data = np.empty((cst.count(), cst[0].stats.npts))
    # y grid
    y = np.zeros((len(cst),))
    # x coords
    times = cst[0].times()
    for ii, ctr in enumerate(cst):
        if ctr.stats.az >= 180:
            data[ii, :] = np.flip(ctr.data)
        else:
            data[ii, :] = ctr.data
        y[ii] = ctr.stats.dist
    ds = plt.pcolormesh(
        times, y, data, shading='auto', cmap=cmap, vmin=vmin,
        vmax=vmax)
    if plot_reference_v:
        yref = np.array([ax.get_ylim()[1], 0, ax.get_ylim()[1]])
        [plt.plot(
            [-yref[0]/rv, 0, yref[0]/rv], yref, linewidth=.5, zorder=99,
            label=f'{rv}'+r'$\frac{km}{s}$') for rv in ref_v]
        plt.legend()
    plt.colorbar(
        ds, label='Correlation Coefficient', shrink=.6,
        orientation='horizontal')
    plt.yticks(np.linspace(0, ctr.stats.dist, 10, dtype=int))
    ax.set_ylabel(r"Distance [km]")
    return times



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 29th March 2021 12:54:05 pm
Last Modified: Wednesday, 5th April 2023 10:32:09 am
'''
from typing import List, Tuple
import logging
import re
import warnings

import numpy as np
from obspy import Inventory, Stream, Trace, UTCDateTime
from obspy.core import Stats, AttribDict


log_lvl = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR}


def trace_calc_az_baz_dist(stats1: Stats, stats2: Stats) -> Tuple[
        float, float, float]:
    """
    Return azimuth, back azimhut and distance between tr1 and tr2
    This funtions calculates the azimut, back azimut and distance between tr1
    and tr2 if both have geo information in their stats dictonary.
    Required fields are:
    ``tr.stats.sac.stla`` and ``tr.stats.sac.stlo``

    :type stats1: :class:`~obspy.core.Stats`
    :param stats1: First trace to account
    :type stats2: :class:`~obspy.core.Stats`
    :param stats2: Second trace to account

    :rtype: float
    :return: **az**: Azimuth angle between tr1 and tr2
    :rtype: float
    :return: **baz**: Back-azimuth angle between tr1 and tr2
    :rtype: float
    :return: **dist**: Distance between tr1 and tr2 in m
    """

    if not isinstance(stats1, (Stats, AttribDict)):
        raise TypeError("stats1 must be an obspy Stats object.")

    if not isinstance(stats2, (Stats, AttribDict)):
        raise TypeError("stats2 must be an obspy Stats object.")

    try:
        from obspy.geodetics import gps2dist_azimuth
    except ImportError:
        print("Missed obspy funciton gps2dist_azimuth")
        print("Update obspy.")
        return

    dist, az, baz = gps2dist_azimuth(
        stats1['stla'], stats1['stlo'], stats2['stla'], stats2['stlo'])

    return az, baz, dist


def filter_stat_dist(inv1: Inventory, inv2: Inventory, thres: float) -> bool:
    """
    Very simple function to check whether to stations are closer than thres
    to each other.

    :param inv1: Inventory of station 1
    :type inv1: Inventory
    :param inv2: Inventory of station 2
    :type inv2: Inventory
    :param thres: Threshold distance in m
    :type thres: float
    :return: True if closer (or equal) than thres, False if not.
    :rtype: bool
    """
    return inv_calc_az_baz_dist(inv1, inv2)[-1] <= thres


def inv_calc_az_baz_dist(inv1: Inventory, inv2: Inventory) -> Tuple[
        float, float, float]:
    """ Return azimuth, back azimuth and distance between stat1 and stat2


    :type tr1: :class:`~obspy.core.inventory.Inventory`
    :param tr1: First trace to account
    :type tr2: :class:`~obspy.core.inventory.Inventory`
    :param tr2: Second trace to account

    :rtype: float
    :return: **az**: Azimuth angle between stat1 and stat2
    :rtype: float
    :return: **baz**: Back-azimuth angle between stat2 and stat2
    :rtype: float
    :return: **dist**: Distance between stat1 and stat2 in m
    """

    if not isinstance(inv1, Inventory):
        raise TypeError("inv1 must be an obspy Inventory.")

    if not isinstance(inv2, Inventory):
        raise TypeError("inv2 must be an obspy Inventory.")

    try:
        from obspy.geodetics import gps2dist_azimuth
    except ImportError:
        print("Missing obspy funciton gps2dist_azimuth")
        print("Update obspy.")
        return

    dist, az, baz = gps2dist_azimuth(
        inv1[0][0].latitude, inv1[0][0].longitude, inv2[0][0].latitude,
        inv2[0][0].longitude)

    return az, baz, dist


def resample_or_decimate(
    data: Trace or Stream, sampling_rate_new: int,
        filter=True) -> Stream or Trace:
    """Decimates the data if the desired new sampling rate allows to do so.
    Else the signal will be interpolated (a lot slower).

    :param data: Stream to be resampled.
    :type data: Stream
    :param sampling_rate_new: The desired new sampling rate
    :type sampling_rate_new: int
    :return: The resampled stream
    :rtype: Stream
    """
    if isinstance(data, Stream):
        if not len(data):
            return data
        sr = data[0].stats.sampling_rate
        srl = [tr.stats.sampling_rate for tr in data]
        if len(set(srl)) != 1:
            # differing sampling rates in stream
            for tr in data:
                try:
                    tr = resample_or_decimate(tr, sampling_rate_new, filter)
                except ValueError:
                    warnings.warn(
                        f'Trace {tr} not downsampled. Sampling rate is lower'
                        + ' than requested sampling rate.')
            return data
    elif isinstance(data, Trace):
        sr = data.stats.sampling_rate
    else:
        raise TypeError('Data has to be an obspy Stream or Trace.')

    srn = sampling_rate_new
    if srn > sr:
        raise ValueError('New sampling rate greater than old. This function \
            is only intended for downsampling.')
    elif srn == sr:
        return data

    # Chosen this filter design as it's exactly the same as
    # obspy.Stream.decimate uses
    # Decimation factor
    factor = float(sr)/float(srn)
    if filter and factor <= 16:
        freq = sr * 0.5 / factor
        data.filter('lowpass_cheby_2', freq=freq, maxorder=12)
    elif filter:
        # Use a different filter
        freq = sr * 0.45 / factor
        data.filter('lowpass_cheby_2', freq=freq, maxorder=12)
    if sr/srn == sr//srn:
        return data.decimate(int(sr//srn), no_filter=True)
    else:
        return data.resample(srn)


def trim_stream_delta(
        st: Stream, start: float, end: float, *args, **kwargs) -> Stream:
    """
    Cut all traces to starttime+start and endtime-end. *args* and *kwargs* will
    be passed to :func:`~obspy.Stream.trim`

    :param st: Input Stream
    :type st: Stream
    :param start: Delta to add to old starttime (in seconds)
    :type start: float
    :param end: Delta to remove from old endtime (in seconds)
    :type end: float
    :return: The trimmed stream
    :rtype: Stream

    .. note::

        This operation is performed in place on the actual data arrays.
        The raw data will no longer be accessible afterwards. To keep your
        original data, use copy() to create a copy of your stream object.

    """
    for tr in st:
        tr = trim_trace_delta(tr, start, end, *args, **kwargs)
    return st


def trim_trace_delta(
        tr: Trace, start: float, end: float, *args, **kwargs) -> Trace:
    """
    Cut all traces to starttime+start and endtime-end. *args* and *kwargs* will
    be passed to :func:`~obspy.Trace.trim`.

    :param st: Input Trace
    :type st: Trace
    :param start: Delta to add to old starttime (in seconds)
    :type start: float
    :param end: Delta to remove from old endtime (in seconds)
    :type end: float
    :return: The trimmed trace
    :rtype: Trace

    .. note::

        This operation is performed in place on the actual data arrays.
        The raw data will no longer be accessible afterwards. To keep your
        original data, use copy() to create a copy of your stream object.

    """
    return tr.trim(
        starttime=tr.stats.starttime+start, endtime=tr.stats.endtime-end,
        *args, **kwargs)


# Time keys
t_keys = ['starttime', 'endtime', 'corr_start', 'corr_end']
# No stats, keys that are not in stats but attributes of the respective objects
no_stats = [
    'corr', 'value', 'sim_mat', 'second_axis', 'method_array', 'vt_array',
    'data', 'tw_len', 'tw_start', 'freq_min', 'freq_max', 'subdir',
    'plot_vel_change', 'start_date', 'end_date', 'win_len', 'date_inc',
    'sides', 'compute_tt', 'rayleigh_wave_velocity', 'stretch_range',
    'stretch_steps', 'dt_ref', 'preprocessing', 'postprocessing']


def save_header_to_np_array(stats: Stats) -> dict:
    """
    Converts an obspy header to a format that allows it to be saved in an
    npz file (i.e., several gzipped npy files)

    :param stats: input Header
    :type stats: Stats
    :return: Dictionary, whose keys are the names of the arrays and the values
        the arrays themselves. Can be fed into ``np.savez`` as `**kwargs`
    :rtype: dict
    """
    array_dict = {}
    for k in stats:
        if k in t_keys:
            array_dict[k] = convert_utc_to_timestamp(stats[k])
        else:
            array_dict[k] = np.array([stats[k]])
    return array_dict


def load_header_from_np_array(array_dict: dict) -> dict:
    """
    Takes the *dictionary-like* return-value of `np.load` and converts the
    corresponding keywords into an obspy header.

    :param array_dict: Return value of `np.load`
    :type array_dict: dict
    :return: The obspy header as dictionary
    :rtype: dict
    """
    d = {}
    for k in array_dict:
        if k in no_stats or re.match('reftr', k):
            continue
        elif k in t_keys:
            d[k] = convert_timestamp_to_utcdt(array_dict[k])
        else:
            try:
                d[k] = array_dict[k][0]
            except IndexError:
                warnings.warn(
                    f'Key {k} could not be loaded into the header.')
    return d


def convert_utc_to_timestamp(
        utcdt: UTCDateTime or List[UTCDateTime]) -> np.ndarray:
    """
    Converts :class:`obspy.core.utcdatetime.UTCDateTime` objects to floats.

    :param utcdt: The input times, either a list of utcdatetimes or one
        utcdatetime
    :type utcdt: UTCDateTimeorList[UTCDateTime]
    :return: A numpy array of timestamps
    :rtype: np.ndarray
    """
    if isinstance(utcdt, UTCDateTime):
        utcdt = [utcdt]
    timestamp = np.array([t.timestamp for t in utcdt])
    return timestamp


def convert_timestamp_to_utcdt(timestamp: np.ndarray) -> List[UTCDateTime]:
    """
    Converts a numpy array holding timestamps (i.e., floats) to a list of
    UTCDateTime objects

    :param timestamp: numpy array holding timestamps
    :type timestamp: np.ndarray
    :return: a list of UTCDateTime objects
    :rtype: List[UTCDateTime]
    """
    timestamp = list(timestamp)
    for ii, t in enumerate(timestamp):
        timestamp[ii] = UTCDateTime(t)
    if len(timestamp) == 1:
        timestamp = timestamp[0]
    return timestamp


def get_valid_traces(st: Stream):
    """Return only valid traces of a stream.

    Remove traces that are 100% masked from a stream. This happens when
    a masked trace is trimmed within a gap. The function works in place.

    :type st: obspy.Stream
    :param st: stream to work on

    """
    for tr in st:
        if isinstance(tr.data, np.ma.MaskedArray):
            if tr.data.mask.all():
                st.remove(tr)
    return


def discard_short_traces(st: Stream, length: float):
    """
    Discard all traces from stream that are shorter than length.

    :param st: inputer obspy Stream
    :type st: Stream
    :param length: Maxixmum Length that should be discarded (in seconds).
    :type length: float

    .. note:: Action is performed in place.
    """
    for tr in st:
        if tr.stats.npts/tr.stats.sampling_rate <= length:
            st.remove(tr)
            logging.debug(f'Discarding short Trace {tr}.')
    return


def nan_moving_av(
        data: np.ndarray, win_half_len: int, axis: int = -1) -> np.ndarray:
    """
    Returns a filtered version of data, disregarding the nans.
    Moving mean window length is win_half_len*2+1.

    :param data: Array to be filtered
    :type data: np.ndarray
    :param win_half_len: Half length of the boxcar filter (len = halflen*2+1)
    :type win_half_len: int
    :param axis: Axis to filter along, defaults to -1
    :type axis: int, optional
    :return: The filtered array
    :rtype: np.ndarray
    """
    # Swap axes, so we can work on queried axis
    dataswap = data.swapaxes(0, axis)
    data_smooth = np.empty_like(dataswap)
    for ii in range(dataswap.shape[0]):
        start = ii - win_half_len
        if start < 0:
            start = 0
        # weighted average
        data_smooth[ii] = np.nanmean(
            dataswap[start:ii+win_half_len+1], axis=0)
    return data_smooth.swapaxes(0, axis)


def stream_require_dtype(st: Stream, dtype: type) -> Stream:
    """
    Often it might make sense to change the data type of seismic data before
    saving or broadcasting (e.g., to save memory). This function allows
    to do so

    :param st: input Stream
    :type st: Stream
    :param dtype: desired datatype, e.g. np.float32
    :type dtype: type
    :return: Stream with data in new dtype
    :rtype: Stream

    .. note:: This operation is performed in place.
    """
    for tr in st:
        tr.data = np.require(tr.data, dtype)


def correct_polarity(st: Stream, inv: Inventory):
    """
    Sometimes the polarity of a seismometer's vertical component is reversed.
    This simple functions flips the polarity if this should be the case.

    .. note::
        This function acts on the data in-place.

    :param st: input stream. If there is no Z-component, nothing will happen.
    :type st: Stream
    :param inv: Inventory holding the orientation information.
    :type inv: Inventory
    """
    st_z = st.select(component='Z')
    for tr in st_z:
        dip = inv.get_orientation(
            tr.id, datetime=tr.stats.starttime)['dip']
        if dip > 0:
            tr.data *= -1



'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 5th July 2021 02:44:13 pm
Last Modified: Monday, 30th January 2023 02:26:32 pm
'''
from obspy.core.util import AttribDict
from obspy import UTCDateTime


class CorrStats(AttribDict):
    """
    From Obspy, but with the difference that some items can be lists and that
    corr_start and corr_end are introduced, some other differences only for
    correlations.

    A container for additional header information of a ObsPy Trace object.

    A ``Stats`` object may contain all header information (also known as meta
    data) of a :class:`~seismic.correlate.stream.CorrTrace` object. Those
    headers may be accessed or modified either in the dictionary style or
    directly via a corresponding attribute. There are various default
    attributes which are required by every waveform import and export modules
    within ObsPy such as :mod:`obspy.io.mseed`.

    :type header: dict or :class:`~obspy.core.trace.Stats`, optional
    :param header: Dictionary containing meta information of a single
        :class:`~obspy.core.trace.Trace` object. Possible keywords are
        summarized in the following `Default Attributes`_ section.

    .. rubric:: Basic Usage

    >>> stats = CorrStats()
    >>> stats.network = 'BW'
    >>> print(stats['network'])
    BW
    >>> stats['station'] = 'MANZ'
    >>> print(stats.station)
    MANZ

    .. rubric:: _`Default Attributes`

    ``sampling_rate`` : float, optional
        Sampling rate in hertz (default value is 1.0).
    ``delta`` : float, optional
        Sample distance in seconds (default value is 1.0).
    ``calib`` : float, optional
        Calibration factor (default value is 1.0).
    ``npts`` : int, optional
        Number of sample points (default value is 0, which implies that no data
        is present).
    ``network`` : string, optional
        Network code (default is an empty string).
    ``location`` : string, optional
        Location code (default is an empty string).
    ``station`` : string, optional
        Station code (default is an empty string).
    ``channel`` : string, optional
        Channel code (default is an empty string).
    ``starttime`` : :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        Date and time of the first data sample used for correlation in UTC
        (default value is "1970-01-01T00:00:00.0Z").
    ``endtime`` : :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        Date and time of the last data sample used for correlation in UTC
        (default value is "1970-01-01T00:00:00.0Z").
    ``corr_start``: :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        Date and time of the first data sample used for correlation in UTC
        (default value is "1970-01-01T00:00:00.0Z").
    ``corr_end``: :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
        Date and time of the last data sample used for correlation in UTC
        (default value is "1970-01-01T00:00:00.0Z").
    ``start_lag``: float, optional
        Lag of the first sample in seconds (usually negative)
    ``end_lag``: float, optional
        Lag of the last sample in seconds.

    .. rubric:: Notes

    (1) The attributes ``sampling_rate`` and ``delta`` are linked to each
        other. If one of the attributes is modified the other will be
        recalculated.

        >>> stats = Stats()
        >>> stats.sampling_rate
        1.0
        >>> stats.delta = 0.005
        >>> stats.sampling_rate
        200.0

    (2) The attributes ``start_lag``, ``npts``, ``sampling_rate`` and ``delta``
        are monitored and used to automatically calculate the ``end_lag``.

        >>> stats = Stats()
        >>> stats.npts = 61
        >>> stats.delta = 1.0
        >>> stats.start_lag = -30
        >>> stats.end_lag
        30
        >>> stats.delta = 0.5
        >>> stats.end_lag
        0

    (3) The attribute ``endtime``, ``end_lag``, and ``starttime`` are
        read only and cannot be modified. ``starttime`` and ``endtime`` are
        just simply aliases of ``corr_start`` and ``corr_end``.

        >>> stats = Stats()
        >>> stats.endtime = UTCDateTime(2009, 1, 1, 12, 0, 0)
        Traceback (most recent call last):
        ...
        AttributeError: Attribute "endtime" in Stats object is read only!
        >>> stats['endtime'] = UTCDateTime(2009, 1, 1, 12, 0, 0)
        Traceback (most recent call last):
        ...
        AttributeError: Attribute "endtime" in Stats object is read only!

    (4)
        The attribute ``npts`` will be automatically updated from the
        :class:`~seismic.correlate.stream.CorrTrace` object.

        >>> trace = CorrTrace()
        >>> trace.stats.npts
        0
        >>> trace.data = np.array([1, 2, 3, 4])
        >>> trace.stats.npts
        4

    (5)
        The attribute ``component`` can be used to get or set the component,
        i.e. the last character of the ``channel`` attribute.

        >>> stats = Stats()
        >>> stats.channel = 'HHZ'
        >>> stats.component  # doctest: +SKIP
        'Z'
        >>> stats.component = 'L'
        >>> stats.channel  # doctest: +SKIP
        'HHL'

    """
    # set of read only attrs
    readonly = ['endtime', 'end_lag', 'starttime']
    # default values
    defaults = {
        'sampling_rate': 1.0,
        'delta': 1.0,
        'starttime': UTCDateTime(0),
        'endtime': UTCDateTime(0),
        'corr_start': UTCDateTime(0),
        'corr_end': UTCDateTime(0),
        'start_lag': 0,
        'end_lag': 0,
        'npts': 0,
        'calib': 1.0,
        'network': '',
        'station': '',
        'location': '',
        'channel': '',
    }
    # keys which need to refresh derived values
    _refresh_keys = {
        'delta', 'sampling_rate', 'corr_start', 'corr_end', 'npts',
        'start_lag'}
    # dict of required types for certain attrs
    _types = {
        'network': (str),
        'station': (str),
    }

    def __init__(self, header={}):
        """
        """
        super(CorrStats, self).__init__(header)

    def __setitem__(self, key, value):
        """
        """
        if key in self._refresh_keys:
            # ensure correct data type
            if key == 'delta':
                key = 'sampling_rate'
                try:
                    value = 1.0 / float(value)
                except ZeroDivisionError:
                    value = 0.0
            elif key == 'sampling_rate':
                value = float(value)
            elif key == 'start_lag':
                value = float(value)
            elif key == 'npts':
                if not isinstance(value, int):
                    value = int(value)
            # set current key
            super(CorrStats, self).__setitem__(key, value)
            # set derived value: delta
            try:
                delta = 1.0 / float(self.sampling_rate)
            except ZeroDivisionError:
                delta = 0
            self.__dict__['delta'] = delta
            # set derived value: endtime
            if self.npts == 0:
                timediff = 0
            else:
                timediff = float(self.npts - 1) * delta
            self.__dict__['end_lag'] = self.start_lag + timediff
            self.__dict__['endtime'] = self.corr_end
            self.__dict__['starttime'] = self.corr_start
            return
        if key == 'component':
            key = 'channel'
            value = str(value)
            if len(value) != 3:
                msg = 'Component must be set with three characters, e.g. E-Z'
                raise ValueError(msg)
            a, b = self.channel.split('-')
            an, bn = value.split('-')
            value = a[:-1] + an + '-' + b[:-1] + bn
        # all other keys
        if isinstance(value, dict):
            super(CorrStats, self).__setitem__(key, AttribDict(value))
        else:
            super(CorrStats, self).__setitem__(key, value)

    __setattr__ = __setitem__

    def __getitem__(self, key, default=None):
        """
        """
        if key == 'component':
            ch = super(CorrStats, self).__getitem__('channel', default)
            a, b = ch.split('-')
            return a[-1]+'-'+b[-1]
        elif key == 'id':
            net = super(CorrStats, self).__getitem__('network', default)
            sta = super(CorrStats, self).__getitem__('station', default)
            ch = super(CorrStats, self).__getitem__('channel', default)
            return f'{net}.{sta}.{ch}'
        else:
            return super(CorrStats, self).__getitem__(key, default)

    def __str__(self):
        """
        Return better readable string representation of Stats object.
        """
        priorized_keys = ['network', 'station', 'location', 'channel',
                          'corr_start', 'corr_end', 'start_lag', 'end_lag',
                          'sampling_rate', 'delta', 'npts', 'calib']
        return self._pretty_str(priorized_keys)

    def _repr_pretty_(self, p, cycle):
        p.text(str(self))



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 5th October 2021 11:50:22 am
Last Modified: Thursday, 21st October 2021 02:44:05 pm
'''

from logging import warn
import os
import fnmatch
from glob import glob
from typing import List, Tuple
from datetime import datetime

import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np

from seismic.monitor.dv import read_dv
from seismic.plot.plot_utils import set_mpl_params


def plot_multiple_dv(
    indir: str, only_mean: bool = False, title: str = None,
    outfile: str = None, fmt: str = None, dpi: int = 300, legend: bool = False,
    plot_median: bool = False, ylim: Tuple[float, float] = None,
    xlim: Tuple[datetime, datetime] = None, statfilter: List[str] = None,
        plt_abs: bool = False):
    """
    Plots several Velocity variations in one single plot

    :param indir: directory in which the .npz DV files are located.
    :type indir: str
    :param only_mean: plot only the average of one station instead of all
        of its components. **The averages have to be computed before**.
        Defaults to false.
    :type only_mean: bool, optional
    :param title: Title for the figure, defaults to None
    :type title: str, optional
    :param outfile: File to save figure to, defaults to None
    :type outfile: str, optional
    :param fmt: Figure format, defaults to None
    :type fmt: str, optional
    :param dpi: DPI for pixel-based formats, defaults to 300
    :type dpi: int, optional
    :param legend: Should a legend be plotted, defaults to False
    :type legend: bool, optional
    :param plot_median: Plot the median of all dataset as a black line.
        Defaults to False.
    :type plot_median: bool, optional.
    :param ylim: y limit, defaults to None
    :type ylim: Tuple[float, float], optional
    :param xlim: x limit, defaults to None
    :type xlim: Tuple[datetime, datetime], optional
    :param statfilter: Only plot data from the station combinations with the
        following codes. Given in the form
        ['net0-net0.sta0-stat1.ch0-ch1', ...]. Note that wildcards are allowed.
        Defaults to None.
    :type statfilter: List[str]
    :param plt_abs: plot absolute values. Defaults to False
    :type plt_abs: bool, optional
    """
    set_mpl_params()
    if only_mean:
        pat = os.path.join(indir, '*.av-*.npz')
    else:
        pat = os.path.join(indir, '*.npz')
    statcodes = []
    vals = []
    infiles = glob(pat)
    if statfilter is not None:
        statfilter = ['*%s*' % filt for filt in statfilter]
        infiles_new = []
        for filt in statfilter:
            infiles_new.extend(fnmatch.filter(infiles, filt))
        infiles = infiles_new
    for fi in infiles:
        try:
            dv = read_dv(fi)
        except Exception:
            warn('Corrupt file %s discovered...skipping.' % fi)
        val = -dv.value
        if plt_abs:
            val = abs(val)
        vals.append(val)
        rtime = [utcdt.datetime for utcdt in dv.stats['corr_start']]
        plt.plot(rtime, val, '.', markersize=.25)
        vals.append(val)
        ax = plt.gca()
        statcodes.append(dv.stats.station)
    ax.xaxis.set_major_locator(mpl.dates.AutoDateLocator())
    if plot_median:
        mean = np.nanmedian(vals, axis=0)
        plt.plot(rtime, mean, 'k')
        statcodes.append('median')

    ax.xaxis.set_major_formatter(mpl.dates.DateFormatter('%d %h'))
    plt.xticks(rotation='vertical')
    plt.ylabel('dv/v')
    if legend:
        plt.legend(statcodes)
    if xlim:
        plt.xlim(xlim)
    if ylim:
        plt.ylim(ylim)
    if title:
        plt.title(title)
    if outfile:
        plt.tight_layout()
        if fmt == 'pdf' or fmt == 'svg':
            plt.savefig('%s.%s' % (outfile, fmt), format=fmt)
        else:
            plt.savefig('%s.%s' % (outfile, fmt), format=fmt, dpi=dpi)
    else:
        plt.show()



'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Friday, 16th July 2021 02:30:02 pm
Last Modified: Monday, 21st August 2023 12:22:18 pm
'''

from datetime import datetime
from typing import Tuple, List
import os
import locale

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib import dates as mdates
import numpy as np
from obspy import UTCDateTime

from seismic.plot.plot_utils import set_mpl_params


def plot_fancy_dv(
    dv, xlim: Tuple[datetime, datetime] = None,
    ylim: Tuple[float, float] = None, return_ax: bool = False,
        title: str = None, dateformat: str = '%d %b %y', *args, **kwargs):
    """ Prettier than the technical plot, but less informative"""
    fig = plt.figure(figsize=(12, 8))

    val = -dv.value[~np.isnan(dv.corr)]*100
    corr_starts = np.array(dv.stats.corr_start)
    t_real = [t.datetime for t in corr_starts[~np.isnan(dv.corr)]]

    # plot dv/v
    map = plt.scatter(
        t_real, val, c=dv.corr[~np.isnan(dv.corr)], cmap='inferno_r', s=22)

    # Correct format of X-Axis
    ax = plt.gca()
    ax.set_ylim(ylim)
    plt.ylabel(r'$\frac{dv}{v}$ [%]')
    plt.grid(True, axis='y')
    locale.setlocale(locale.LC_ALL, "en_GB.utf8")
    ax.xaxis.set_major_formatter(mdates.DateFormatter(dateformat))
    plt.xticks(rotation=25)

    fig.colorbar(
        map, orientation='horizontal', pad=0.1, shrink=.6,
        label='Correlation Coefficient', anchor=(0.5, .8))
    plt.title(title)
    if xlim is not None:
        plt.xlim(xlim)
    else:
        plt.xlim((min(t_real), max(t_real)))
    if return_ax:
        return fig, ax


def plot_technical_dv(
    dv, save_dir='.', figure_file_name=None, mark_time=None,
    normalize_simmat=False, sim_mat_Clim=[], figsize=(9, 11), dpi=72,
    ylim: Tuple[float, float] = None, xlim: Tuple[datetime, datetime] = None,
    title: str = None, plot_scatter: bool = False,
    return_ax: bool = False, *args,
        **kwargs) -> Tuple[plt.figure, List[plt.axis]]:
    """
    Plot a :class:`seismic.monitor.dv.DV` object
    The produced figure is saved in `save_dir` that, if necessary, it is
    created.
    It is also possible to pass a "special" time value `mark_time` that will be
    represented in the `dv/v` and `corr` plot as a vertical line; It can be
    a string (i.e. YYYY-MM-DD) or directly a :class:`~datetime.datetime`
    object.
    if the `dv` dictionary also contains a 'comb_mseedid' keyword, its `value`
    (i.e. MUST be a string) will be reported in the title.
    In case of the chosen filename exist in the `save_dir` directory, a prefix
    _<n> with n:0..+Inf, is added.
    The aspect of the plot may change depending on the matplotlib version. The
    recommended one is matplotlib 1.1.1

    :type dv: dict
    :param dv: velocity change estimate dictionary as output by
        :class:`~miic.core.stretch_mod.multi_ref_vchange_and_align` and
        successively "extended" to contain also the timing in the form
        {'time': time_vect} where `time_vect` is a :class:`~numpy.ndarray` of
        :class:`~datetime.datetime` objects.
    :type save_dir: string
    :param save_dir: Directory where to save the produced figure. It is created
        if it doesn't exist.
    :type figure_file_name: string
    :param figure_file_name: filename to use for saving the figure. If None
        the figure is displayed in interactive mode.
    :type mark_time: string or :class:`~datetime.datetime` object
    :param mark_time: It is a "special" time location that will be represented
        in the `dv/v` and `corr` plot as a vertical line.
    :type normalize_simmat: Bool
    :param normalize_simmat: if True the simmat will be normalized to a maximum
        of 1. Defaults to False
    :type sim_mat_Clim: 2 element array_like
    :param sim_mat_Clim: if non-empty it set the color scale limits of the
        similarity matrix image
    :param ylim: Limits for the stretch axis. Defaults to None
    :type ylim: Tuple[float, float], optional
    :param return_ax: Return plt.figure and list of axes. Defaults to False.
        This overwrites any choice to save the figure.
    :type return_ax: bool, optional
    :returns: If `return_ax` is set to True it returns fig and axes.
    """
    set_mpl_params()

    if sim_mat_Clim and len(sim_mat_Clim) != 2:
        raise ValueError('Sim_mat_Clim has to be a list or Tuple of length 2.')

    if figure_file_name:
        os.makedirs(save_dir, exist_ok=True)

    # Create a unique filename if TraitsUI-default is given
    if figure_file_name == 'plot_default':
        fname = figure_file_name + '_change.png'
        exist = os.path.isfile(os.path.join(save_dir, fname))
        i = 0
        while exist:
            fname = "%s_%i" % (figure_file_name, i)
            exist = os.path.isfile(os.path.join(save_dir,
                                                fname + '_change.png'))
            i += 1
        figure_file_name = fname

    # Extract the data from the dictionary

    value_type = dv.value_type
    method = dv.method
    corr = dv.corr
    dt = dv.value
    sim_mat = dv.sim_mat
    stretch_vect = dv.second_axis
    stats = dv.stats

    rtime = np.array(
        [utcdt.datetime for utcdt in stats['corr_start']],
        dtype=np.datetime64)

    # normalize simmat if requested
    if normalize_simmat:
        sim_mat = sim_mat/np.tile(
            np.max(sim_mat, axis=1), (sim_mat.shape[1], 1)).T

    stretching_amount = np.max(stretch_vect)

    # Adapt plot details in agreement with the type of dictionary that
    # has been passed

    if (value_type == 'stretch') and (method == 'single_ref'):

        tit = "Single reference dv/v"
        # Find order of magnitude of max strech
        oom = -int(np.floor(np.log10(stretch_vect.max()/5)))
        dv_tick_delta = round(stretch_vect.max()/5, oom)
        dv_y_label = "dv/v"
        # plotting velocity requires to flip the stretching axis
    elif (value_type == 'stretch') and (method == 'multi_ref'):

        # Find order of magnitude of max strech
        oom = -int(np.floor(np.log10(stretch_vect.max()/5)))
        tit = "Multi reference dv/v"
        dv_tick_delta = round(stretch_vect.max()/5, oom)
        dv_y_label = "dv/v"
        # plotting velocity requires to flip the stretching axis
    elif (value_type == 'shift') and (method == 'time_shift'):

        tit = "Time shift"
        dv_tick_delta = 5
        dv_y_label = "Time Shift (sample)"
    elif (value_type == 'shift') and (method == 'absolute_shift'):
        tit = "Time shift"
        dv_tick_delta = (
            np.max(dv.second_axis)
            - np.min(dv.second_axis))/5
        dv_y_label = "Time Shift [s]"
    else:
        raise ValueError(f"Unknown dv type, {value_type}!")

    f = plt.figure(figsize=figsize, dpi=dpi)

    if dv.n_stat is not None and plot_scatter:
        gs = mpl.gridspec.GridSpec(4, 1, height_ratios=[12, 4, 4, 1])
    else:
        gs = mpl.gridspec.GridSpec(3, 1, height_ratios=[3, 1, 1])

    ax1 = f.add_subplot(gs[0])
    imh = plt.imshow(
        np.flipud(sim_mat.T).astype(float), interpolation='none',
        aspect='auto')

    # plotting value is way easier now
    plt.plot(-dv.value, 'b.')

    # Set extent so we can treat the axes properly (mainly y)
    imh.set_extent((0, sim_mat.shape[0], stretch_vect[-1], stretch_vect[0]))

    ###
    if xlim:
        xlim = (np.datetime64(xlim[0]), np.datetime64(xlim[1]))
        xl0 = np.argmin(abs(rtime-xlim[0]))
        xl1 = np.argmin(abs(rtime-xlim[1]))
        ax1.set_xlim(xl0, xl1+1)
        plt.xlim(xl0, xl1+1)
    else:
        ax1.set_xlim(0, sim_mat.shape[0])
        plt.xlim(0, sim_mat.shape[0])

    if value_type == 'stretch':
        ax1.invert_yaxis()
    if ylim:
        plt.ylim(ylim)
    # ###
    if sim_mat_Clim:
        imh.set_clim(sim_mat_Clim[0], sim_mat_Clim[1])

    plt.gca().get_xaxis().set_visible(False)

    comb_mseedid = '%s.%s.%s' % (
        stats['network'], stats['station'],  # stats['location'],
        stats['channel'])
    if title:
        tit = title
    else:
        tit = "%s estimate (%s)" % (tit, comb_mseedid)

    ax1.set_title(tit)
    ax1.yaxis.set_ticks_position('right')
    ax1.yaxis.set_label_position('left')
    ax1.set_xticklabels([])
    ax1.set_ylabel(dv_y_label)

    ax2 = f.add_subplot(gs[1])
    if plot_scatter:
        # reshape so we can plot a histogram
        histt = np.array([t.timestamp for t in stats['corr_start']])
        histt = np.tile(histt, dv.stretches.shape[0])
        histcorrs = np.reshape(dv.corrs, -1)
        histstretches = np.reshape(-dv.stretches, -1)
        n_bins = np.flip(np.array(dv.sim_mat.shape))//10
        # remove nans
        nanmask = ~np.isnan(histcorrs)
        histt = histt[nanmask]
        histcorrs = histcorrs[nanmask]
        histstretches = histstretches[nanmask]
        # create histograms
        H_c, xedges_c, yedges_c = np.histogram2d(histt, histcorrs, bins=n_bins)
        H_v, xedges_v, yedges_v = np.histogram2d(
            histt, histstretches, bins=n_bins)
        xedges_c = [UTCDateTime(xe).datetime for xe in xedges_c]
        xedges_v = [UTCDateTime(xe).datetime for xe in xedges_v]
        plt.pcolor(xedges_v, yedges_v, H_v.T, cmap='binary')
    plt.plot(rtime, -dt, '.', markersize=3)
    if xlim:
        plt.xlim(xlim[0], xlim[1])
    else:
        plt.xlim([rtime[0], rtime[-1]])
    if ylim:
        plt.ylim(ylim)
    else:
        plt.ylim((-stretching_amount, stretching_amount))
    if mark_time and not (
            np.all(rtime < mark_time) and np.all(rtime > mark_time)):
        plt.axvline(mark_time, lw=1, color='r')
    ax2.yaxis.set_ticks_position('left')
    ax2.yaxis.set_label_position('right')
    ax2.yaxis.label.set_rotation(270)
    ax2.yaxis.set_label_coords(1.03, 0.5)
    ax2.set_ylabel(dv_y_label)
    ax2.yaxis.set_major_locator(plt.MultipleLocator(dv_tick_delta))
    ax2.yaxis.grid(True, 'major', linewidth=1)
    ax2.xaxis.grid(True, 'major', linewidth=1)
    ax2.set_xticklabels([])

    ax3 = f.add_subplot(gs[2])
    if plot_scatter:
        plt.pcolor(xedges_c, yedges_c, H_c.T, cmap='binary')
    plt.plot(rtime, corr, '.', markersize=3)

    if xlim:
        plt.xlim(xlim[0], xlim[1])
    else:
        plt.xlim([rtime[0], rtime[-1]])
    ax3.yaxis.set_ticks_position('right')
    ax3.set_ylabel("Correlation")
    plt.ylim((0, 1))
    if mark_time and not (
            np.all(rtime < mark_time) and np.all(rtime > mark_time)):
        plt.axvline(mark_time, lw=1, color='r')
    ax3.yaxis.grid(True, 'major', linewidth=1)
    ax3.xaxis.grid(True, 'major', linewidth=1)
    ax3.yaxis.set_major_locator(plt.MultipleLocator(0.2))
    # Plot number of stations
    if dv.n_stat is not None and plot_scatter:
        ax4 = f.add_subplot(gs[3])
        plt.fill_between(
            rtime, 0, dv.n_stat, interpolate=False)
        plt.setp(ax4.get_xticklabels(), rotation=45, ha='right')
        ax3.set_xticklabels([])
        ax4.yaxis.set_label_position('right')
        ax4.set_ylabel("N")
        plt.ylim(0, int(dv.n_stat.max()*1.1))
        if xlim:
            plt.xlim(xlim[0], xlim[1])
        else:
            plt.xlim([rtime[0], rtime[-1]])
    else:
        ax4 = None
        plt.setp(ax3.get_xticklabels(), rotation=45, ha='right')

    plt.subplots_adjust(hspace=0, wspace=0)
    if return_ax:
        return f, [ax1, ax2, ax3, ax4]
    if figure_file_name is None:
        plt.show()
    else:
        print('saving to %s' % figure_file_name)
        if figure_file_name.split('.')[-1].lower() not in [
                'png', 'svg', 'pdf', 'jpg']:
            figure_file_name += '.png'
        f.savefig(os.path.join(save_dir, figure_file_name),
                  dpi=dpi)


def plot_dv(style: str, *args, **kwargs):
    if style == 'technical':
        return plot_technical_dv(*args, **kwargs)
    elif style == 'publication':
        return plot_fancy_dv(*args, **kwargs)
    else:
        raise ValueError(f'Unknown style: {style}')



'''
Module for waveform data analysis. Contains spectrogram computation.

:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
   EUROPEAN UNION PUBLIC LICENCE Version 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Wednesday, 21st June 2023 12:22:00 pm
Last Modified: Thursday, 24th August 2023 10:52:44 am
'''
from typing import Iterator
import warnings

from mpi4py import MPI
import numpy as np
from obspy import Stream, Trace
from scipy.signal import welch
from scipy.interpolate import pchip_interpolate

from seismic.utils.miic_utils import resample_or_decimate


def spct_series_welch(
        streams: Iterator[Stream], window_length: int, freqmax: float):
    """
    Computes a spectral time series. Each point in time is computed using the
    welch method. Windows overlap by half the windolength. The input stream can
    contain one or several traces from the same station. Frequency axis is
    logarithmic.

    .. note::
        MPI support since version 0.4.2. Then the result is only
        available on rank 0. All other ranks return None. (to save RAM)

    :param st: Input Stream with data from one station.
    :type st: ~obspy.core.Stream
    :param window_length: window length in seconds for each datapoint in time
    :type window_length: int or float
    :param freqmax: maximum frequency to be considered
    :type freqmax: float
    :return: Arrays containing a frequency and time series and the spectral
        series.
    :rtype: np.ndarray
    """
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    psize = comm.Get_size()
    specl = []
    # List of actually available times
    t = []
    for ii, st in enumerate(streams):
        # Find out which stream to process
        if ii % psize != rank:
            continue
        try:
            # Don't preprocess masked values
            tr = Stream(
                [preprocess(tr, freqmax) for tr in st.split()]).merge()[0]
        except IndexError:
            warnings.warn('No data in stream for this time step.')
            continue
        except Exception as e:
            warnings.warn(
                'Error while preprocessing stream. Skipping... Message: '
                f'{e}')
            continue
        for wintr in tr.slide(window_length=window_length, step=window_length):
            try:
                f, S = welch(wintr.data, fs=wintr.stats.sampling_rate)
                # interpolate onto a logarithmic frequency space
                # 512 points of resolution in f direction hardcoded for now
                f2 = np.logspace(-3, np.log10(f.max()), 512)
                S2 = pchip_interpolate(f, S, f2)
                specl.append(S2)
                t.append(wintr.stats.starttime)
            except Exception as e:
                warnings.warn(
                    'Error while computing welch spectrum for time window: '
                    f'{wintr.stats.starttime}. Skipping... Message: {e}')
                continue
    # Gather all the data
    specl = comm.gather(specl, root=0)
    t = comm.gather(t, root=0)
    if rank == 0:
        # Flatten the list
        specl = [item for sublist in specl for item in sublist]
        t = [item for sublist in t for item in sublist]
        # Sort from values in t
        specl = [x for _, x in sorted(zip(t, specl))]
        t = sorted(t)

        # Convert to numpy array
        S = np.array(specl)
        t = np.array(t)
        try:
            return f2, t, S.T
        except UnboundLocalError:
            raise ValueError(
                'A spectrum could not be computed. Check earlier log messages')
    else:
        return None, None, None


def preprocess(tr: Trace, freqmax: float):
    """
    Some very basic preprocessing on the string in order to plot the spectral
    series. Does the following steps:
    *1. Remove station response*
    *2. Detrend*
    *3. Decimate if sampling rate>50*
    *4. highpass filter with corner period of 300s.*
    :param st: The input Stream, should only contain Traces from one station.
    :type st: ~obspy.core.Stream
    :return: The output stream and station inventory object
    :rtype: ~obspy.core.Stream and ~obspy.core.Inventory
    """
    # Downsample to make computations faster
    resample_or_decimate(tr, freqmax*2)
    tr.remove_response()
    # Detrend
    tr.detrend(type='linear')

    # highpass filter
    tr.filter('highpass', freq=0.01)

    return tr
